import time

import scphylo as scp
from scphylo.tl.solver.booster._dependencies import prepare_dependencies
from scphylo.tl.solver.booster._reconstruct_big_tree import reconstruct_big_tree
from scphylo.tl.solver.booster._subsamples import subsampling


def booster(
    df_input,
    alpha,
    beta,
    solver="SCITE",
    sample_on="muts",
    sample_size=10,
    n_samples=10,
    begin_index=0,
    n_jobs=10,
    dep_weight=50,
    time_limit=120,
    n_iterations=500000,
    subsample_dir=None,
    disable_tqdm=False,
    no_subsampling=False,
    no_dependencies=False,
    no_reconstruction=False,
):
    """Trisicell-Boost solver.

    For more details of available tools that work on binary matrices, read
    :cite:`ReviewBinary`.

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        input noisy dataframe
    alpha : :obj:`float`
        false positive rate
    beta : :obj:`float`
        false negative rate
    solver : :obj:`str`, optional
        which tool is boosted {"SCITE", "PhISCS"}, by default "SCITE"
    sample_on : :obj:`str`, optional
        on which dimension is subsampled {"muts", "cells"}, by default "muts"
    sample_size : :obj:`int`, optional
        number of subsampled mutations or cells depends on `sample_on`, by default 10
    n_samples : :obj:`int`, optional
        number of samples, by default 10
    begin_index : :obj:`int`, optional
        start index of intermediate file names, by default 0
    n_jobs : :obj:`int`, optional
        number of jobs, by default 10
    dep_weight : :obj:`int`, optional
        weight multiplier, by default 50
    time_limit : :obj:`int`, optional
        time out needed for PhISCS running on each instance, by default 120
    n_iterations : :obj:`int`, optional
        number of iterations needed for SCITE running, by default 500000
    subsample_dir : :obj:`str`, optional
        for keeping the intermediate subsamples CFMatrices, by default None
    disable_tqdm : :obj:`bool`, optional
        disable progress bar, by default False
    no_subsampling : :obj:`bool`, optional
        subsampling (step 1/3) gets off, by default False
    no_dependencies : :obj:`bool`, optional
        dependencies calculation (step 2/3) gets off, by default False
    no_reconstruction : :obj:`bool`, optional
        reconstruction of big tree (step 3/3) gets off, by default False

    Returns
    -------
    :class:`pandas.DataFrame`
        A conflict-free matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1) and absence (0).

    See Also
    --------
    :func:`scphylo.tl.scite`.
    :func:`scphylo.tl.phiscsb`
    """
    if subsample_dir is not None:
        tmpdir = scp.ul.mkdir(subsample_dir)
    else:
        # tmpdir = scp.ul.tmpdirsys(suffix=".booster")
        # tmpdir = tmpdir.name
        tmpdir = scp.ul.tmpdir(suffix=".booster")

    n_muts = df_input.shape[1]
    if sample_size is None:
        sample_size = int(dep_weight * (n_muts**2) / (sample_size**2))

    detail = {}
    df_input.index.name = "cellIDxmutID"

    s_time = time.time()
    # subsampling matrices and solving them
    if not no_subsampling:
        subsampling(
            df_input,
            alpha=alpha,
            beta=beta,
            solver=solver,
            sample_on=sample_on,
            sample_size=sample_size,
            n_samples=n_samples,
            begin_sample=begin_index,
            n_jobs=n_jobs,
            time_limit=time_limit,
            n_iterations=n_iterations,
            tmpdir=tmpdir,
            disable_tqdm=disable_tqdm,
        )

    # preparing dependencies file
    if not no_dependencies:
        max_num_submatrices = int(dep_weight * (n_muts**2) / (sample_size**2))
        prepare_dependencies(
            df_input.columns,
            tmpdir,
            f"{tmpdir}/_booster.dependencies",
            max_num_submatrices,
            disable_tqdm,
        )

    # building the final CFMatrix
    if not no_reconstruction:
        scp.io.write(df_input, f"{tmpdir}/_input.SC")
        detail["TREE_SCORE"] = reconstruct_big_tree(
            f"{tmpdir}/_booster.dependencies",
            f"{tmpdir}/_input.SC",
            alpha,
            beta,
            f"{tmpdir}/_booster",
            disable_tqdm,
        )
        df_output = scp.io.read(
            f"{tmpdir}/_booster.dnc.CFMatrix",
        )
        df_output = df_output.loc[df_input.index, df_input.columns]
    else:
        df_output = None
    e_time = time.time()
    running_time = e_time - s_time

    if subsample_dir is None:
        scp.ul.cleanup(tmpdir)

    if df_output is not None:
        scp.ul.stat(df_input, df_output, alpha, beta, running_time)
        for k, v in detail.items():
            scp.logg.info(f"{k}: {v}")

    return df_output



import os

import click

import scphylo as scp


@click.command(short_help="Run OncoNEM.")
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "alpha",
    required=True,
    type=float,
)
@click.argument(
    "beta",
    required=True,
    type=float,
)
def onconem(genotype_file, alpha, beta):
    """OncoNEM.

    Inferring tumor evolution from single-cell sequencing data :cite:`OncoNEM`.

    scphylo solver onconem input.SC 0.0001 0.1
    """
    outfile = os.path.splitext(genotype_file)[0]

    scp.settings.verbosity = "info"
    scp.settings.logfile = f"{outfile}.onconem.log"

    df_in = scp.io.read(genotype_file)
    df_out = scp.tl.onconem(df_in, alpha=alpha, beta=beta)
    scp.io.write(df_out, f"{outfile}.onconem.CFMatrix")

    return None



import gzip
import subprocess

import click
import pandas as pd

import scphylo as scp
from scphylo.ul._servers import cmd, write_cmds_get_main


@click.command(short_help="Run STAR.")
@click.argument(
    "indir",
    required=True,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
)
@click.argument(
    "outdir",
    required=True,
    type=click.Path(file_okay=False, dir_okay=True, readable=True, resolve_path=True),
)
@click.argument(
    "ref",
    required=True,
    type=click.Choice(scp.settings.refs),
)
@click.option(
    "--time",
    default=["2:00:00", "20:00:00"],
    multiple=True,
    show_default=True,
    help="Time List [indexing, mapping].",
)
@click.option(
    "--mem",
    default=["50", "90"],
    multiple=True,
    show_default=True,
    help="Memory List [indexing, mapping].",
)
@click.option(
    "--max_multimapping",
    default="10",
    type=str,
    show_default=True,
    help="`outFilterMultimapNmax` parameter in STAR.",
)
@click.option(
    "--afterok",
    default=None,
    type=str,
    show_default=True,
    help="Afterok.",
)
@click.option(
    "--is_pdx",
    default=False,
    type=bool,
    show_default=True,
    is_flag=True,
    help="Is the the PDX.",
)
def star(indir, outdir, ref, time, mem, max_multimapping, afterok, is_pdx):
    """Run STAR.

    scphylo caller star /path/to/in/dir /path/to/out/dir hg19|hg38|mm10 --is_pdx

    FastQ files (*.fastq.gz) --> BAM files (*.mapped.bam)

    if --is_pdx: FastQ files (*.fastq.gz) --> FastQ files (*.fastq.gz)
    """
    scp.ul.mkdir(outdir)
    if ref == "hg19":
        config = scp.settings.hg19
    elif ref == "mm10":
        config = scp.settings.mm10
    elif ref == "hg38":
        config = scp.settings.hg38
    else:
        config = None

    def get_readlength(filename):
        readlength = None
        with gzip.open(filename, "rb") as fin:
            i = 0
            for line in fin:
                readlength = len(line.strip()) - 1
                if i == 1:
                    break
                i += 1
        return readlength

    def step1(afterok):
        def get_command():
            cmds = ""
            cmds += cmd([f"mkdir -p {outdir}/_indexing/star1"])
            cmds += cmd([f"module load {scp.settings.tools['star']}"])
            cmds += cmd(
                [
                    "STAR",
                    "--runMode genomeGenerate",
                    f"--genomeDir {outdir}/_indexing/star1",
                    f"--genomeFastaFiles {config['ref']}",
                    f"--sjdbGTFfile {config['annot']}",
                    f"--sjdbOverhang {readlength}",
                    "--runThreadN 32",
                ]
            )
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        cmds_tmp = pd.DataFrame()
        cmds_tmp["cmd"] = [get_command()]
        cmdmain = write_cmds_get_main(
            cmds_tmp,
            "star-1of5",
            time[0],
            mem[0],
            None,
            32,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    def step2(afterok):
        def get_command(sample, is_paired):
            if is_paired:
                infqs = [f"{indir}/{sample}_{i}.fastq.gz" for i in range(1, 3)]
            else:
                infqs = [f"{indir}/{sample}.fastq.gz"]
            cmds = ""
            cmds += cmd([f"module load {scp.settings.tools['star']}"])
            cmds += cmd(
                [
                    "STAR",
                    "--runMode alignReads",
                    f"--genomeDir {outdir}/_indexing/star1",
                    f"--sjdbGTFfile {config['annot']}",
                    f"--outFilterMultimapNmax {max_multimapping}",
                    "--outSAMunmapped None",
                    "--quantMode TranscriptomeSAM GeneCounts",
                    "--runThreadN 1",
                    f"--sjdbOverhang {readlength}",
                    "--readFilesCommand zcat",
                    f"--outFileNamePrefix {outdir}/{sample}_",
                    "--readFilesIn",
                    f"{' '.join(infqs)}",
                ]
            )
            cmds += cmd(
                [
                    "rm -rf",
                    f"{outdir}/{sample}_Aligned.out.sam",
                ]
            )
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        df_cmds["cmd"] = df_cmds.apply(
            lambda x: get_command(x["sample"], x["is_paired"]), axis=1
        )
        cmdmain = write_cmds_get_main(
            df_cmds,
            "star-2of5",
            time[1],
            mem[1],
            None,
            1,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    def step3(afterok):
        def get_command():
            files = " ".join(
                [f"{outdir}/{s}_SJ.out.tab" for s in df_cmds["sample"].values]
            )
            cmds = ""
            cmds += cmd([f"mkdir -p {outdir}/_indexing/star2"])
            cmds += cmd([f"module load {scp.settings.tools['star']}"])
            cmds += cmd(
                [
                    "STAR",
                    "--runMode genomeGenerate",
                    f"--genomeDir {outdir}/_indexing/star2",
                    f"--genomeFastaFiles {config['ref']}",
                    f"--sjdbGTFfile {config['annot']}",
                    f"--sjdbFileChrStartEnd {files}",
                    f"--sjdbOverhang {readlength}",
                    "--runThreadN 32",
                ]
            )
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        cmds_tmp = pd.DataFrame()
        cmds_tmp["cmd"] = [get_command()]
        cmdmain = write_cmds_get_main(
            cmds_tmp,
            "star-3of5",
            time[0],
            mem[0],
            None,
            32,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    def step4(afterok):
        def get_command(sample, is_paired):
            if is_paired:
                infqs = [f"{indir}/{sample}_{i}.fastq.gz" for i in range(1, 3)]
            else:
                infqs = [f"{indir}/{sample}.fastq.gz"]
            cmds = ""
            cmds += cmd([f"module load {scp.settings.tools['star']}"])
            tmp = [
                "STAR",
                "--runMode alignReads",
                f"--genomeDir {outdir}/_indexing/star2",
                "--readFilesCommand zcat",
                "--readFilesIn",
                f"{' '.join(infqs)}",
                f"--outFileNamePrefix {outdir}/{sample}_",
                "--limitBAMsortRAM 30000000000",
                "--outSAMtype BAM SortedByCoordinate",
                f"--sjdbGTFfile {config['annot']}",
                f"--outFilterMultimapNmax {max_multimapping}",
                "--outSAMunmapped None",
                "--quantMode TranscriptomeSAM GeneCounts",
                "--runThreadN 1",
                f"--sjdbOverhang {readlength}",
            ]
            if is_pdx:
                tmp += ["--outReadsUnmapped Fastx"]
            cmds += cmd(tmp)
            cmds += cmd(
                [
                    "rm -rf",
                    f"{outdir}/{sample}_Aligned.out.sam",
                    f"{outdir}/{sample}_Log.progress.out",
                    f"{outdir}/{sample}_Log.out",
                    f"{outdir}/{sample}_ReadsPerGene.out.tab",
                    f"{outdir}/{sample}_SJ.out.tab",
                    f"{outdir}/{sample}__STARgenome",
                ]
            )
            cmds += cmd(
                [
                    "mv",
                    f"{outdir}/{sample}_Aligned.sortedByCoord.out.bam",
                    f"{outdir}/{sample}.mapped.bam",
                ]
            )
            cmds += cmd(
                [
                    "mv",
                    f"{outdir}/{sample}_Aligned.toTranscriptome.out.bam",
                    f"{outdir}/{sample}.transcript.bam",
                ]
            )
            cmds += cmd(
                [
                    "mv",
                    f"{outdir}/{sample}_Log.final.out",
                    f"{outdir}/{sample}.star.log",
                ]
            )
            if is_pdx:
                cmds += cmd(
                    [
                        "mv",
                        f"{outdir}/{sample}_Unmapped.out.mate1",
                        f"{outdir}/{sample}_1.fastq",
                    ]
                )
                cmds += cmd(
                    [
                        "mv",
                        f"{outdir}/{sample}_Unmapped.out.mate2",
                        f"{outdir}/{sample}_2.fastq",
                    ]
                )
                cmds += cmd(
                    [
                        "gzip",
                        f"{outdir}/{sample}_1.fastq",
                    ]
                )
                cmds += cmd(
                    [
                        "gzip",
                        f"{outdir}/{sample}_2.fastq",
                    ]
                )
                cmds += cmd(
                    [
                        "rm -rf",
                        f"{outdir}/{sample}.mapped.bam",
                        f"{outdir}/{sample}.transcript.bam",
                    ]
                )
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        df_cmds["cmd"] = df_cmds.apply(
            lambda x: get_command(x["sample"], x["is_paired"]), axis=1
        )
        cmdmain = write_cmds_get_main(
            df_cmds,
            "star-4of5",
            time[1],
            mem[1],
            None,
            1,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    def step5(afterok):
        def get_command():
            cmds = ""
            cmds += cmd(
                [
                    f"python {scp.ul.get_file('scphylo.ul/star.py')}",
                    f"{outdir}",
                ]
            )
            cmds += cmd(
                [
                    "rm -rf",
                    f"{outdir}/*.star.log",
                ]
            )
            cmds += cmd(["echo Done!"], islast=True)
            return cmds

        cmds_tmp = pd.DataFrame()
        cmds_tmp["cmd"] = [get_command()]
        cmdmain = write_cmds_get_main(
            cmds_tmp,
            "star-5of5",
            time[0],
            mem[0],
            None,
            2,
            scp.settings.tools["email"],
            f"{outdir}/_tmp",
            afterok,
        )
        return cmdmain

    temp = scp.ul.is_paired(indir)
    if temp[0]["is_paired"]:
        readlength = get_readlength(f"{indir}/{temp[0]['sample']}_1.fastq.gz")
    else:
        readlength = get_readlength(f"{indir}/{temp[0]['sample']}.fastq.gz")
    df_cmds = pd.DataFrame(temp)
    if df_cmds.shape[0] != df_cmds["sample"].nunique():
        scp.logg.error("Samples are not unique!")

    cmd1 = step1(afterok)
    code1 = subprocess.getoutput(cmd1)
    cmd2 = step2(code1)
    code2 = subprocess.getoutput(cmd2)
    cmd3 = step3(code2)
    code3 = subprocess.getoutput(cmd3)
    cmd4 = step4(code3)
    code4 = subprocess.getoutput(cmd4)
    cmd5 = step5(code4)
    code5 = subprocess.getoutput(cmd5)
    scp.logg.info(code5)

    return None



import glob
import os

import click
import pandas as pd

import scphylo as scp
from scphylo.ul._servers import cmd, write_cmds_get_main


@click.command(short_help="Run deFuse.")
@click.argument(
    "indir",
    required=True,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
)
@click.argument(
    "outdir",
    required=True,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
)
@click.argument(
    "ref",
    required=True,
    type=click.Choice(scp.settings.refs),
)
@click.option(
    "--time",
    default="0-02:00:00",
    type=str,
    show_default=True,
    help="Time.",
)
@click.option(
    "--mem",
    default="50",
    type=str,
    show_default=True,
    help="Memory.",
)
@click.option(
    "--n_threads",
    default="16",
    type=str,
    show_default=True,
    help="Number of Threads.",
)
@click.option(
    "--afterok",
    default=None,
    type=str,
    show_default=True,
    help="Afterok.",
)
def defuse(indir, outdir, ref, time, mem, n_threads, afterok):
    """Run deFuse.

    scphylo caller defuse ...

    FastQ files (*.fastq) -> deFuse files
    """
    if ref == "hg19":
        config = scp.settings.hg19
    elif ref == "mm10":
        config = scp.settings.mm10
    elif ref == "hg38":
        config = scp.settings.hg38
    else:
        config = None

    def cmds(sample):
        cmds = ""
        cmds += cmd([f"module load {scp.settings.tools['defuse']}"])
        cmds += cmd([f"mkdir -p {outdir}/{sample}"])
        cmds += cmd(
            [
                "defuse.pl",
                f"-c {config['defuse_config']}",
                f"-o {outdir}/{sample}",
                f"-d {config['defuse_db']}",
                f"-1 {indir}/{sample}_R1_001.fastq",
                f"-2 {indir}/{sample}_R2_001.fastq",
                "-s direct",
                "-p $SLURM_CPUS_PER_TASK",
            ]
        )
        cmds += cmd(["echo Done!"], islast=True)
        return cmds

    df = pd.DataFrame()
    temp = []
    for x in glob.glob(f"{indir}/*_R1_001.fastq"):
        temp.append(x.split("/")[-1].replace("_R1_001.fastq", ""))
    df["sample"] = temp
    df["cmd"] = df.apply(lambda x: cmds(x["sample"]), axis=1)
    scp.logg.info(f"Number of samples: {df.shape[0]}")

    cmdmain = write_cmds_get_main(
        df,
        "defuse",
        time,
        mem,
        None,
        n_threads,
        scp.settings.tools["email"],
        f"{outdir}/_tmp",
        afterok,
    )
    os.system(cmdmain)

    return None



#!/usr/bin/env python
import glob
import os
import sys

import anndata as ad
import pandas as pd

outdir = sys.argv[1]
adata = ad.AnnData()
files = glob.glob(outdir + "/*.transcript.bam")
kinds = {"expected_count": "count", "TPM": "tpm", "FPKM": "fpkm"}
for kind in kinds.keys():
    i = 0
    expr = pd.DataFrame()
    for file in files:
        file = file[: -len(".transcript.bam")] + ".genes.results"
        df = pd.read_csv(file, sep="\t")
        name = os.path.basename(file)[: -len(".genes.results")]
        df = df[["gene_id", "expected_count", "TPM", "FPKM"]]
        df = df.set_index(["gene_id"])
        if i == 0:
            expr = pd.DataFrame(index=df.index)
        expr[name] = df[kind]
        i += 1
    expr = expr.transpose()
    expr = expr.sort_index(ascending=True)
    if kind == "expected_count":
        adata = ad.AnnData(expr)
    else:
        adata.layers[kind] = expr.values
adata.write(outdir + "/_expression.h5ad.gz", compression="gzip")



import networkx as nx
import numpy as np
import pandas as pd

import scphylo as scp


def to_tree(df):
    """Convert a conflict-free matrix to a tree object.

    This function converts a conflict-free matrix to a tree object in which
    nodes are labled with cells and edges are lables with mutations. The root is
    labled by 'root'. Mutations are seperated by `.graph['splitter_mut']` and cells
    are seperated by `.graph['splitter_cell']`. Those mutations that are not present
    in any cell are stored in `.graph['become_germline']`. Mutations happed once
    during the evolution so there is no repetitive mutation.

    Parameters
    ----------
    df : :class:`pandas.DataFrame`
        A genotype dataframe in which rows are cells and columns are mutations.
        Note that this dataframe must be conflict-free.

    Returns
    -------
    :class:`networkx.DiGraph`
        A perfect phylogenetic tree.
    """
    if not scp.ul.is_conflict_free_gusfield(df):
        scp.logg.error("The input is not conflict-free!")

    def _contains(col1, col2):
        for i in range(len(col1)):
            if not col1[i] >= col2[i]:
                return False
        return True

    tree = nx.DiGraph()
    tree.graph["data"] = df
    tree.graph["splitter_mut"] = "\n"
    tree.graph["splitter_cell"] = "\n"
    tree.graph["become_germline"] = df.columns[(df == 0).all(axis=0)]

    matrix = df.values
    names_mut = list(df.columns)

    i = 0
    while i < matrix.shape[1]:
        j = i + 1
        while j < matrix.shape[1]:
            if np.array_equal(matrix[:, i], matrix[:, j]):
                matrix = np.delete(matrix, j, 1)
                x = names_mut.pop(j)
                names_mut[i] += tree.graph["splitter_mut"] + x
                j -= 1
            j += 1
        i += 1

    # rows = matrix.shape[0]
    cols = matrix.shape[1]
    dimensions = np.sum(matrix, axis=0)
    indices = np.argsort(dimensions)
    dimensions = np.sort(dimensions)
    names_mut = [names_mut[indices[i]] for i in range(cols)]

    tree.add_node(cols)
    tree.add_node(cols - 1)
    tree.add_edge(cols, cols - 1, label=names_mut[cols - 1])
    node_mud = {}
    node_mud[names_mut[cols - 1]] = cols - 1

    i = cols - 2
    while i >= 0:
        if dimensions[i] == 0:
            break
        attached = False
        for j in range(i + 1, cols):
            if _contains(matrix[:, indices[j]], matrix[:, indices[i]]):
                tree.add_node(i)
                tree.add_edge(node_mud[names_mut[j]], i, label=names_mut[i])
                node_mud[names_mut[i]] = i
                attached = True
                break
        if not attached:
            tree.add_node(i)
            tree.add_edge(cols, i, label=names_mut[i])
            node_mud[names_mut[i]] = i
        i -= 1

    tumor_cells = []
    clusters = {cols: "root"}
    for node in tree:
        if node == cols:
            tree.nodes[node]["label"] = "root"
            continue
        untilnow_mut = []
        sp = nx.shortest_path(tree, cols, node)
        for i in range(len(sp) - 1):
            untilnow_mut += tree.get_edge_data(sp[i], sp[i + 1])["label"].split(
                tree.graph["splitter_mut"]
            )
        untilnow_cell = df.loc[
            (df[untilnow_mut] == 1).all(axis=1)
            & (df[[x for x in df.columns if x not in untilnow_mut]] == 0).all(axis=1)
        ].index
        if len(untilnow_cell) > 0:
            clusters[node] = f"{tree.graph['splitter_cell'].join(untilnow_cell)}"
            tumor_cells += list(
                y for y in tree.graph["splitter_cell"].join(untilnow_cell)
            )
        else:
            clusters[node] = "––"

        tree.nodes[node]["label"] = clusters[node]

    tree.graph["normal_cells"] = df[df.sum(axis=1) == 0].index
    tree.graph["root_id"] = cols

    i = 1
    for k, v in clusters.items():
        if v == "––":
            clusters[k] = i * "––"
            i += 1
    return tree


def to_cfmatrix(tree):
    """Convert phylogenetic tree to conflict-free matrix.

    This function converts a phylogenetic tree in which mutations
    are at edges and cells are at nodes to a conflict-free matrix
    where rows are cells, columns are mutations and each entry is
    either zero or one representing the absence or presence of the
    mutation in the cell.

    Parameters
    ----------
    tree : :class:`networkx.DiGraph`
        The phylogenetic tree.

    Returns
    -------
    :class:`pandas.DataFrame`
        The conflict-free matrix.
    """
    mutations = []
    cells = []
    for _, v, lable in tree.edges(data=True):
        mutations += lable["label"].split(tree.graph["splitter_mut"])
        if "––" not in tree.nodes[v]["label"]:
            cells += tree.nodes[v]["label"].split(tree.graph["splitter_cell"])
    df = pd.DataFrame(0, index=cells, columns=mutations)
    root = scp.ul.root_id(tree)
    leaves = [x for x in tree.nodes if tree.out_degree(x) == 0]
    for leaf in leaves:
        nodes = nx.dijkstra_path(tree, root, leaf)
        mut = []
        for first, second in zip(nodes, nodes[1:]):
            mut += tree.edges[(first, second)]["label"].split(
                tree.graph["splitter_mut"]
            )
            if len(mut) != len(set(mut)):
                print(tree.edges[(first, second)]["label"])
                return mut
            if "––" not in tree.nodes[second]["label"]:
                cell = tree.nodes[second]["label"].split(tree.graph["splitter_cell"])
                df.loc[cell, mut] = 1
    return df


def to_mtree(tree):
    """Convert the phylogenetic tree to mutation tree.

    Parameters
    ----------
    tree : :class:`networkx.DiGraph`
        The phylogenetic tree in which cells are in nodes and
        mutations are at edges.

    Returns
    -------
    :class:`networkx.DiGraph`
        The mutation tree in which mutations are in nodes.
    """
    tree2 = nx.DiGraph()
    for u, v, l in tree.edges.data("label"):
        if tree.in_degree(u) == 0:
            tree2.add_node(u, label="root")
        muts = l.split(tree.graph["splitter_mut"])
        tree2.add_node(v, label=muts)
        tree2.add_edge(u, v)
    return tree2


def _to_newick(tree):
    def _children(at):
        return list(n for n in tree.neighbors(at))

    root = scp.ul.root_id(tree)

    def _newick_recursive(node_id):
        node_ids = _children(node_id)
        if len(node_ids) == 0:
            cells = tree.nodes[node_id]["label"].split(tree.graph["splitter_cell"])
            return "(" + ",".join(cells) + f")Node{node_id+1}"
        elif len(node_ids) > 0:
            cells = tree.nodes[node_id]["label"].split(tree.graph["splitter_cell"])
            if not ("––" in cells or "root" in cells):
                return (
                    "("
                    + ",".join(cells)
                    + ","
                    + ",".join([_newick_recursive(node_id) for node_id in node_ids])
                    + f")Node{node_id+1}"
                )
            else:
                return (
                    "("
                    + ",".join([_newick_recursive(node_id) for node_id in node_ids])
                    + f")Node{node_id+1}"
                )
        else:
            return None

    newick = _newick_recursive(root) + ";"
    return newick


def _split_labels(tree1, tree2):
    def _which_muts_must_be_splited(tree):
        muts = []
        rid = scp.ul.root_id(tree)
        for n1 in tree.nodes:
            if n1 != rid:
                if len(tree.nodes[n1]["label"]) >= 2:
                    muts += tree.nodes[n1]["label"]
        return muts

    def _sorted_muts(tree):
        root_guide = scp.ul.root_id(tree)
        guide = [
            y
            for x in nx.algorithms.bfs_tree(tree, root_guide).nodes
            for y in tree.nodes[x]["label"]
        ]
        return guide

    def _splitter_helper(mt, muts_guide):
        root = scp.ul.root_id(mt)
        nodes = list(mt.nodes)
        latest_node = root
        removing = []
        for node in nodes:
            muts = mt.nodes[node]["label"]
            if len(muts) > 1 and node != root:
                parent = list(mt.predecessors(node))[-1]
                if mt.out_degree(node) != 0:
                    children = list(mt.successors(node))
                else:
                    children = -1
                muts = sorted(muts, key=lambda i: muts_guide.index(i))

                removing.append(node)
                for i, x in enumerate(muts):
                    latest_node += 1
                    mt.add_node(latest_node, label=x)
                    if i == 0:
                        mt.add_edge(parent, latest_node)
                    elif i == len(muts) - 1:
                        if children != -1:
                            mt.add_edge(latest_node - 1, latest_node)
                            for child in children:
                                mt.add_edge(latest_node, child)
                        else:
                            mt.add_edge(latest_node - 1, latest_node)
                    else:
                        mt.add_edge(latest_node - 1, latest_node)
            else:
                if node != root:
                    mt.nodes[node]["label"] = mt.nodes[node]["label"][0]

        for node in removing:
            mt.remove_node(node)
        return mt

    muts1 = _which_muts_must_be_splited(tree1)
    muts2 = _which_muts_must_be_splited(tree2)
    muts_c = sorted(np.intersect1d(muts1, muts2))
    guide1 = _sorted_muts(tree1)
    guide2 = _sorted_muts(tree2)
    muts_12 = sorted(np.setdiff1d(muts1, muts2), key=lambda i: guide2.index(i))
    muts_21 = sorted(np.setdiff1d(muts2, muts1), key=lambda i: guide1.index(i))

    mt_tree1 = _splitter_helper(tree1.copy(), muts_c + muts_12)
    mt_tree2 = _splitter_helper(tree2.copy(), muts_c + muts_21)

    return mt_tree1, mt_tree2


def _to_apted(sl_tree):
    def _children(at):
        return list(n for n in sl_tree.neighbors(at))

    def _apted_recursive(node):
        subgs = "{" + sl_tree.nodes[node]["label"]
        nodes = _children(node)
        for node in nodes:
            if len(_children(node)) > 0:
                subgs += _apted_recursive(node)
            else:
                subgs += "{" + sl_tree.nodes[node]["label"] + "}"
        return subgs + "}"

    root = scp.ul.root_id(sl_tree)
    return _apted_recursive(root)


def root_id(tree):
    for x in tree.nodes:
        if tree.in_degree(x) == 0:
            return x
    return None


def partition_cells(tree, node):
    nd = tree.graph["splitter_cell"].join(node)
    cells = []
    for x in list(nx.algorithms.traversal.depth_first_search.dfs_tree(tree, nd).nodes):
        for y in x.split(", "):
            if "–" not in y:
                cells.append(y)
    cells = np.array(cells)
    return cells, np.setdiff1d(tree.graph["data"].index, cells)


def cells_rooted_at(tree, node_id):
    muts = tree.graph["mutation_list"][tree.graph["mutation_list"].Node == node_id]
    if muts.index.shape[0] == 0:
        return np.array([])
    cells = (tree.graph["data"][muts.index] == 1).all(axis=1)
    cells = np.array(tree.graph["data"].loc[cells].index)
    return cells  # , np.setdiff1d(tree.graph["data"].index, cells)


def muts_rooted_at(tree, node_id):
    muts = tree.graph["mutation_list"][tree.graph["mutation_list"].Node == node_id]
    if muts.index.shape[0] == 0:
        return np.array([])

    muts = muts.index.tolist()
    nd = int(node_id.replace("[", "").replace("]", ""))
    paths = nx.algorithms.traversal.depth_first_search.dfs_tree(tree, nd).nodes
    sub_tree = nx.subgraph(tree, paths)
    for _, _, label in sub_tree.edges.data("label"):
        muts += label.split(tree.graph["splitter_mut"])
    return np.array(muts)


def is_leaf(tree, node):
    return "––" not in tree.nodes[node]["label"] and tree.in_degree(node) != 0



import networkx as nx
import numpy as np

import scphylo as scp


def _get_cnt_tree(tree):
    cnt_tree = tree.copy()
    for u, v, muts in cnt_tree.edges.data("label"):
        mutations = muts.split(cnt_tree.graph["splitter_mut"])
        cnt_tree.add_edge(u, v, label=len(mutations), mutations=mutations)
        if cnt_tree.in_degree(v) != 0:
            if "––" not in tree.nodes[v]["label"]:
                cnt_tree.nodes[v]["label"] = cnt_tree.nodes[v]["label"].split(
                    cnt_tree.graph["splitter_cell"]
                )
    return cnt_tree


def _get_edge_set(tree):
    edges = {}
    for e in tree.edges():
        nodes = nx.algorithms.traversal.depth_first_search.dfs_tree(tree, e[1]).nodes
        cells = []
        for n in nodes:
            if "––" not in tree.nodes[n]["label"]:
                cells += tree.nodes[n]["label"]
        edges[tuple(sorted(cells))] = e
    return edges


def consensus(sc1, sc2):
    """Build the consensus tree between two tumor progression trees.

    Parameters
    ----------
    df1 : :class:`pandas.DataFrame`
        First conflict-free matrix.
    df2 : :class:`pandas.DataFrame`
        Second conflict-free matrix.

    Returns
    -------
    :class:`networkx.DiGraph`
        The consensus tree in which cells are in edges and mutations are in edges.
    """
    good_cells1 = sc1.index[(sc1 != 0).sum(axis=1) > 0]
    good_cells2 = sc2.index[(sc2 != 0).sum(axis=1) > 0]
    common_cells = np.intersect1d(good_cells1, good_cells2)
    if len(common_cells) == 0:
        scp.logg.error("No common cells found in two inputs!")
    sc1_p = sc1.loc[common_cells, :]
    sc2_p = sc2.loc[common_cells, :]

    tree1 = scp.ul.to_tree(sc1_p)
    tree2 = scp.ul.to_tree(sc2_p)

    cnt_tree1 = _get_cnt_tree(tree1)
    cnt_tree2 = _get_cnt_tree(tree2)

    edge_set1 = _get_edge_set(cnt_tree1)
    edge_set2 = _get_edge_set(cnt_tree2)

    lists = list(edge_set1.items())
    lists.sort(key=lambda x: len(x[0]))
    final_tree = cnt_tree1.copy()
    cost = 0
    for c, e in lists:
        if c not in edge_set2:
            if final_tree.nodes[e[0]]["label"] == "––":
                label = final_tree.nodes[e[1]]["label"]
            else:
                if final_tree.nodes[e[1]]["label"] == "––":
                    label = final_tree.nodes[e[0]]["label"]
                else:
                    label = (
                        final_tree.nodes[e[0]]["label"]
                        + final_tree.nodes[e[1]]["label"]
                    )
            cost += final_tree.edges[e]["label"]
            final_tree = nx.contracted_nodes(final_tree, e[0], e[1], self_loops=False)
            final_tree.nodes[e[0]]["label"] = label
    scp.logg.info("    ---> total cost:", cost)

    # convert back to the scphylo tree format
    for v in final_tree.nodes:
        if final_tree.in_degree(v) != 0:
            if "––" not in final_tree.nodes[v]["label"]:
                final_tree.nodes[v]["label"] = final_tree.graph["splitter_cell"].join(
                    final_tree.nodes[v]["label"]
                )
    i = 0
    for e, u, _ in final_tree.edges.data("label"):
        final_tree.edges[(e, u)]["label"] = f"m{i}"
        i += 1
    data = scp.ul.to_cfmatrix(final_tree)
    final_tree = scp.ul.to_tree(data)

    return final_tree



import networkx as nx
import numpy as np

import scphylo as scp


def fitch(tree: nx.DiGraph):
    """[summary].

    Parameters
    ----------
    tree : nx.DiGraph
        The tree in which leaves has `'profile'` attribute.

    Returns
    -------
    [type]
        [description]
    """
    tree_bu = _fitch_hartigan_bottom_up(tree)
    tree_td = _fitch_hartigan_top_down(tree_bu)
    tree_fi = _parsimony_cost(tree_td)
    return tree_fi


def _get_consensus(a, b):
    if a == 1 and b == 1:
        return 1
    if a == 0 and b == 0:
        return 0
    if a == 2 and (b == 0 or b == 1):
        return b
    if b == 2 and (a == 0 or a == 1):
        return a
    return 2


_get_consensus = np.vectorize(_get_consensus)


def _fill_child(a, b):
    if b == 2:
        return a
    return b


_fill_child = np.vectorize(_fill_child)


def _set_depth(G, root):
    depth = nx.shortest_path_length(G, root)
    for d in depth.keys():
        G.nodes[d]["depth"] = depth[d]
    return G


def _cut_tree(G, depth):
    nodes = []
    for n in G.nodes:
        if G.nodes[n]["depth"] == depth:
            nodes.append(n)
    return nodes


def _get_max_depth(G, root):
    md = 0
    for n in nx.descendants(G, root):
        if G.nodes[n]["depth"] > md:
            md = G.nodes[n]["depth"]
    return md


def _fitch_hartigan_bottom_up(tree2):
    tree = tree2.copy()
    root = [n for n in tree.nodes() if tree.in_degree(n) == 0][0]
    # leaves = [n for n in tree.nodes() if tree.out_degree(n) == 0]

    tree = _set_depth(tree, root)
    md = _get_max_depth(tree, root)
    d = md

    while d >= 0:
        internal_nodes = _cut_tree(tree, d)
        for i in internal_nodes:
            children = list(tree.successors(i))
            if len(children) == 0:
                continue
            if len(children) == 1:
                tree.nodes[i]["profile"] = tree.nodes[children[0]]["profile"]
            else:
                gc = tree.nodes[children[0]]["profile"]
                for child in children[1:]:
                    gc = _get_consensus(
                        gc,
                        tree.nodes[child]["profile"],
                    )
                tree.nodes[i]["profile"] = gc
        d -= 1

    return tree


def _fitch_hartigan_top_down(tree2):
    tree = tree2.copy()
    root = [n for n in tree.nodes() if tree.in_degree(n) == 0][0]

    tree = _set_depth(tree, root)
    md = _get_max_depth(tree, root)
    tree.nodes[root]["profile"] = [0] * tree.nodes[root]["profile"].shape[0]
    d = 1

    while d <= md:
        internal_nodes = list(_cut_tree(tree, d))
        for i in internal_nodes:
            parent = list(tree.predecessors(i))[0]
            tree.nodes[i]["profile"] = _fill_child(
                tree.nodes[parent]["profile"], tree.nodes[i]["profile"]
            )
        d += 1
    return tree


def _parsimony_cost(tree):
    cost = 0
    for e in tree.edges():
        source = e[0]
        dest = e[1]
        s = np.sum(np.abs(tree.nodes[source]["profile"] - tree.nodes[dest]["profile"]))
        if s > 0:
            cost += s
            tree.edges[e]["event"] = np.where(
                tree.nodes[source]["profile"] != tree.nodes[dest]["profile"]
            )[0]
        else:
            tree.edges[e]["event"] = []
    scp.logg.info(f"cost: {cost}")
    return tree



from subprocess import PIPE, Popen

import numpy as np

import scphylo as scp
from scphylo.external.gpps._nh2lgf import newick_to_edgelist

__author__ = "Simone Ciccolella"
__date__ = "11/30/21"


class Node:
    def __init__(
        self,
        name,
        parent,
        id_node,
        mutation_id,
        loss=False,
        tot_mutations=0,
        gt_build=True,
    ):
        """Node class.

        Parameters
        ----------
        name : [type]
            [description]
        parent : [type]
            [description]
        id_node : [type]
            [description]
        mutation_id : [type]
            [description]
        loss : bool, optional
            [description], by default False
        tot_mutations : int, optional
            [description], by default 0
        gt_build : bool, optional
            [description], by default True
        """
        self.name = name
        self.id_node = id_node
        self.parent = parent
        self.children = []
        self.loss = loss
        self.mut_id = mutation_id
        self.tot_mutations = tot_mutations
        if parent:
            parent.children.append(self)

            if gt_build:
                gt_par_cp = parent.genotype_profile.copy()
                if self.loss:
                    gt_par_cp[mutation_id] -= 1
                else:
                    gt_par_cp[mutation_id] += 1

                self.genotype_profile = gt_par_cp
        else:
            self.genotype_profile = [0 for x in range(tot_mutations)]

    def is_ancestor_of(self, node):
        par = node.parent
        while par:
            if par == self:
                return True
            par = par.parent
        return False

    def print_node_dot(self):
        if self.parent is None:
            print(f'\t"{self.parent.id_node}" -> "{self.id_node}";')
            if "-" not in self.name:
                print(f'\t"{self.id_node}" [label="{self.name}"];')
            else:
                print(
                    '\t"{}" [color=indianred1, style=filled, label="{}"];'.format(
                        self.id_node, self.name[:-1]
                    )
                )

    def print_node_dot_file(self, fout):
        if self.parent is None:
            fout.write(f'\t"{self.parent.id_node}" -> "{self.id_node}";\n')
            if "-" not in self.name:
                fout.write(f'\t"{self.id_node}" [label="{self.name}"];\n')
            else:
                fout.write(
                    '\t"{}" [color=indianred1, style=filled, label="{}"];\n'.format(
                        self.id_node, self.name[:-1]
                    )
                )


def add_edge(start, end):
    start.children.append(end)
    end.parent = start


def __print_tree(node):
    if len(node.children) == 0:
        node.print_node_dot()
    else:
        node.print_node_dot()
        for child in node.children:
            __print_tree(child)


def __print_tree_file(node, fout):
    if len(node.children) == 0:
        node.print_node_dot_file(fout)
    else:
        node.print_node_dot_file(fout)
        for child in node.children:
            __print_tree_file(child, fout)


def print_dot_tree(node):
    print("digraph phylogeny {")
    print(f'\t"{node.id_node}" [label="{node.name}"];')
    __print_tree(node)
    print("}")


def print_dot_tree_file(node, fout):
    fout.write("digraph phylogeny {\n")
    fout.write(f'\t"{node.id_node}" [label="{node.name}"];\n')
    __print_tree_file(node, fout)
    fout.write("}\n")


def __copy_tree_rec(node, cp_parent, nid_dict):
    node_cp = Node(node.name, cp_parent, node.id_node, node.mut_id, loss=node.loss)
    nid_dict[node_cp.id_node] = node_cp
    if len(node.children) == 0:
        return
    for child in node.children:
        __copy_tree_rec(child, node_cp, nid_dict)


def copy_tree(root):
    nid_nodes = {}
    cp_root = Node(
        root.name,
        root.parent,
        root.id_node,
        root.mut_id,
        tot_mutations=root.tot_mutations,
    )
    nid_nodes[root.id_node] = cp_root
    for child in root.children:
        __copy_tree_rec(child, cp_root, nid_nodes)
    return cp_root, nid_nodes


def contains(col1, col2):
    for i in range(len(col1)):
        if not col1[i] >= col2[i]:
            return False
    return True


def build_tree_from_file(ilp_matrix, mutations_names, mutations_ids, tot_mutations):
    executable = scp.ul.executable("gpps_tree", "gpps tree scripts")
    tmpdir = scp.ul.tmpdirsys(suffix=".gpps")
    np.savetxt(
        f"{tmpdir.name}/input.mtx", ilp_matrix.values, delimiter=" ", fmt="%1.0f"
    )

    rb_tree = Popen(["ruby", executable, "-m", f"{tmpdir.name}/input.mtx"], stdout=PIPE)
    stdout, _ = rb_tree.communicate()
    tree = stdout.decode("utf-8").strip()

    tmpdir.cleanup()

    node_dict, edges = newick_to_edgelist(tree)

    building_dictionary = {}
    nid = 0

    # Get germline nid:
    for k in node_dict:
        if node_dict[k] == "germline":
            nid = k

    root = Node("germline", None, nid, -1, tot_mutations=tot_mutations)
    building_dictionary[nid] = root

    for edge in edges:
        e, s = edge

        x = None
        try:
            x = building_dictionary[s]
        except KeyError:
            x_column_index = int(node_dict[s][1:]) - 1
            if "---" in mutations_names[x_column_index]:
                loss = True
            else:
                loss = False
            x = Node(
                mutations_names[x_column_index],
                None,
                s,
                mutations_ids[x_column_index],
                gt_build=False,
                loss=loss,
            )
            building_dictionary[s] = x

        y = None
        try:
            y = building_dictionary[e]
        except KeyError:
            y_column_index = int(node_dict[e][1:]) - 1
            if "---" in mutations_names[y_column_index]:
                loss = True
            else:
                loss = False
            y = Node(
                mutations_names[y_column_index],
                None,
                e,
                mutations_ids[y_column_index],
                gt_build=False,
                loss=loss,
            )
            building_dictionary[e] = y
        add_edge(x, y)

    calculate_genotype_profile_subtree(root, building_dictionary)

    return (root, building_dictionary)


def calculate_genotype_profile_subtree(node, nid_dict):
    if node.parent:
        gt_par_cp = node.parent.genotype_profile.copy()
        if node.loss:
            gt_par_cp[node.mut_id] -= 1
        else:
            gt_par_cp[node.mut_id] += 1
        node.genotype_profile = gt_par_cp

    else:
        # This assumes that the root is correctly initiated at all 0s
        pass
    if len(node.children) == 0:
        return
    for child in node.children:
        calculate_genotype_profile_subtree(child, nid_dict)


def prune_and_reattach(node_prune, node_reattach, nid_dict):
    if node_prune.is_ancestor_of(node_reattach):
        return False
    node_prune.parent.children.remove(node_prune)
    node_prune.parent = node_reattach
    node_reattach.children.append(node_prune)

    check_subtree_losses(node_reattach, nid_dict)

    # rebuild genotype profile of pruned subtree
    calculate_genotype_profile_subtree(node_reattach, nid_dict)

    return True


def is_loss_valid(node, mut_id):
    par = node.parent
    while par:
        if par.mut_id == mut_id:
            return True
        par = par.parent
    return False


def is_already_lost(node, mut_id):
    par = node.parent
    while par:
        if par.loss and par.mut_id == mut_id:
            return True
        par = par.parent
    return False


def delete_node(node, nid_dict):
    parent = node.parent
    # node.parent = None
    parent.children.remove(node)
    for child in node.children:
        child.parent = parent
        parent.children.append(child)
    nid_dict.pop(node.id_node)
    node = None
    return node


def check_subtree_losses(node, nid_dict):
    if node.loss:
        valid = is_loss_valid(node, node.mut_id)
        lost = is_already_lost(node, node.mut_id)

        if not valid or lost:
            delete_node(node, nid_dict)

    if len(node.children) == 0:
        return
    for child in node.children:
        check_subtree_losses(child, nid_dict)


def import_ilp_out(ilp_matrix, k_dollo, mutation_names):
    mut_names = []
    mut_ids = []

    mut_index = 0
    for mut in mutation_names:
        mut_names.append(mut)
        mut_ids.append(mut_index)
        for _ in range(k_dollo):
            mut_names.append("%s---" % mut)
            mut_ids.append(mut_index)
        mut_index += 1

    imported_tree, imported_dict = build_tree_from_file(
        ilp_matrix, mut_names, mut_ids, len(mutation_names)
    )

    return imported_tree, imported_dict



import os
import time

import pandas as pd

import scphylo as scp


def sphyr(
    df_input,
    alpha,
    beta,
    n_restarts=10,
    n_threads=1,
    time_limit=None,
    n_cell_clusters=10,
    n_mut_clusters=15,
):
    """Solving using SPhyR.

    Tumor phylogeny estimation from single-cell sequencing data under loss and error
    :cite:`SPhyR`.

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        Input genotype matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1), absence (0) and missing
        entires (3).
    alpha : :obj:`float`
        False positive error rate.
    beta : :obj:`float`
        False negative error rate.
    n_restarts : :obj:`int`, optional
        Number of threads, by default 10
    n_threads : :obj:`int`, optional
        Number of threads, by default 1
    time_limit : :obj:`int`, optional
        Time limit (in seconds), by default None
    n_cell_clusters : :obj:`int`, optional
        Number of cell clusters, by default 10
    n_mut_clusters : :obj:`int`, optional
        Number of mutation clusters, by default 15

    Returns
    -------
    :class:`pandas.DataFrame`
        A conflict-free matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1) and absence (0).
    """
    executable = scp.ul.executable("sphyr_kDPFC", "SPhyR")

    scp.logg.info(
        f"running SPhyR with alpha={alpha}, beta={beta}, n_restarts={n_restarts}, "
        f"n_threads={n_threads}, time_limit={time_limit}, "
        f"n_cell_clusters={n_cell_clusters}, n_mut_clusters={n_mut_clusters}"
    )

    tmpdir = scp.ul.tmpdirsys(suffix=".sphyr")
    # tmpdir = scp.ul.tmpdir(suffix=".sphyr")

    with open(f"{tmpdir.name}/sphyr.input", "a") as fout:
        fout.write(f"{df_input.shape[0]} #cells\n{df_input.shape[1]} #SNVs\n")
        df_input.replace(3, -1).to_csv(fout, sep=" ", header=None, index=None)
    with open(f"{tmpdir.name}/sphyr.cellnames", "w") as fout:
        fout.write("\n".join(df_input.index) + "\n")
    with open(f"{tmpdir.name}/sphyr.mutnames", "w") as fout:
        fout.write("\n".join(df_input.columns) + "\n")

    cmd = (
        f"{executable} "
        f"{tmpdir.name}/sphyr.input "
        f"-a {alpha} "
        f"-b {beta} "
        f"-N {n_restarts} "
        f"-t {n_threads} "
        f"-lC {n_mut_clusters} "
        f"-lT {n_cell_clusters} "
        f"-T {time_limit if time_limit is not None else -1} "
        "-k 0 "
        f"> {tmpdir.name}/sphyr.output "
        f"2> {tmpdir.name}/sphyr.log"
    )

    s_time = time.time()
    os.system(cmd)
    e_time = time.time()
    running_time = e_time - s_time

    df_output = pd.read_csv(
        f"{tmpdir.name}/sphyr.output",
        sep=" ",
        skiprows=[0, 1],
        header=None,
    )
    df_output.index = pd.read_csv(f"{tmpdir.name}/sphyr.cellnames", header=None)[0]
    df_output.columns = pd.read_csv(f"{tmpdir.name}/sphyr.mutnames", header=None)[0]
    df_output.index.name = "cellIDxmutID"

    tmpdir.cleanup()

    scp.ul.stat(df_input, df_output, alpha, beta, running_time)

    return df_output



#!/usr/bin/env python
import glob
import sys

import pandas as pd

import scphylo as scp


def get_numreads_percmapped(file):
    data = {
        "total_reads": None,
        "uniquely_mapped": None,
        "uniquely_mapped_percent": None,
        "num_splices": None,
        "num_GCAG_splices": None,
        "insertion_length": None,
        "deletion_length": None,
        "unmapped_tooshort_percent": None,
        "avg_mapped_read_length": None,
        "deletion_rate": None,
        "mismatch_rate": None,
        "avg_input_read_length": None,
        "num_ATAC_splices": None,
        "num_annotated_splices": None,
        "num_GTAG_splices": None,
        "multimapped_toomany": None,
        "unmapped_mismatches": None,
        "unmapped_mismatches_percent": None,
        "unmapped_other": None,
        "insertion_rate": None,
        "unmapped_other_percent": None,
        "multimapped_percent": None,
        "multimapped": None,
        "num_noncanonical_splices": None,
        "unmapped_tooshort": None,
        "multimapped_toomany_percent": None,
    }
    with open(file) as fin:
        for line in fin:
            if "Uniquely mapped reads % |" in line:
                data["uniquely_mapped_percent"] = float(
                    line.replace("Uniquely mapped reads % |", "")
                    .strip()
                    .replace("%", "")
                )
            if "Number of splices: Total |" in line:
                data["num_splices"] = int(
                    line.replace("Number of splices: Total |", "").strip()
                )
            if "Number of splices: GC/AG |" in line:
                data["num_GCAG_splices"] = int(
                    line.replace("Number of splices: GC/AG |", "").strip()
                )
            if "Insertion average length |" in line:
                data["insertion_length"] = float(
                    line.replace("Insertion average length |", "")
                    .strip()
                    .replace("%", "")
                )
            if "Deletion average length |" in line:
                data["deletion_length"] = float(
                    line.replace("Deletion average length |", "")
                    .strip()
                    .replace("%", "")
                )
            if "% of reads unmapped: too short |" in line:
                data["unmapped_tooshort_percent"] = float(
                    line.replace("% of reads unmapped: too short |", "")
                    .strip()
                    .replace("%", "")
                )
            if "Average mapped length |" in line:
                data["avg_mapped_read_length"] = float(
                    line.replace("Average mapped length |", "").strip().replace("%", "")
                )
            if "Deletion rate per base |" in line:
                data["deletion_rate"] = float(
                    line.replace("Deletion rate per base |", "")
                    .strip()
                    .replace("%", "")
                )
            if "Mismatch rate per base, % |" in line:
                data["mismatch_rate"] = float(
                    line.replace("Mismatch rate per base, % |", "")
                    .strip()
                    .replace("%", "")
                )
            if "Average input read length |" in line:
                data["avg_input_read_length"] = float(
                    line.replace("Average input read length |", "")
                    .strip()
                    .replace("%", "")
                )
            if "Number of splices: AT/AC |" in line:
                data["num_ATAC_splices"] = int(
                    line.replace("Number of splices: AT/AC |", "").strip()
                )
            if "Number of splices: Annotated (sjdb) |" in line:
                data["num_annotated_splices"] = int(
                    line.replace("Number of splices: Annotated (sjdb) |", "").strip()
                )
            if "Number of splices: GT/AG |" in line:
                data["num_GTAG_splices"] = int(
                    line.replace("Number of splices: GT/AG |", "").strip()
                )
            if "Uniquely mapped reads number |" in line:
                data["uniquely_mapped"] = int(
                    line.replace("Uniquely mapped reads number |", "").strip()
                )
            if "Number of reads mapped to too many loci |" in line:
                data["multimapped_toomany"] = int(
                    line.replace(
                        "Number of reads mapped to too many loci |", ""
                    ).strip()
                )
            if "Number of reads unmapped: too many mismatches |" in line:
                data["unmapped_mismatches"] = int(
                    line.replace(
                        "Number of reads unmapped: too many mismatches |", ""
                    ).strip()
                )
            if "% of reads unmapped: too many mismatches |" in line:
                data["unmapped_mismatches_percent"] = float(
                    line.replace("% of reads unmapped: too many mismatches |", "")
                    .strip()
                    .replace("%", "")
                )
            if "Number of input reads |" in line:
                data["total_reads"] = int(
                    line.replace("Number of input reads |", "").strip()
                )
            if "Number of reads unmapped: other |" in line:
                data["unmapped_other"] = int(
                    line.replace("Number of reads unmapped: other |", "").strip()
                )
            if "Insertion rate per base |" in line:
                data["insertion_rate"] = float(
                    line.replace("Insertion rate per base |", "")
                    .strip()
                    .replace("%", "")
                )
            if "% of reads unmapped: other |" in line:
                data["unmapped_other_percent"] = float(
                    line.replace("% of reads unmapped: other |", "")
                    .strip()
                    .replace("%", "")
                )
            if "% of reads mapped to multiple loci |" in line:
                data["multimapped_percent"] = float(
                    line.replace("% of reads mapped to multiple loci |", "")
                    .strip()
                    .replace("%", "")
                )
            if "Number of reads mapped to multiple loci |" in line:
                data["multimapped"] = int(
                    line.replace(
                        "Number of reads mapped to multiple loci |", ""
                    ).strip()
                )
            if "Number of splices: Non-canonical |" in line:
                data["num_noncanonical_splices"] = int(
                    line.replace("Number of splices: Non-canonical |", "").strip()
                )
            if "Number of reads unmapped: too short |" in line:
                data["unmapped_tooshort"] = float(
                    line.replace("Number of reads unmapped: too short |", "")
                    .strip()
                    .replace("%", "")
                )
            if "% of reads mapped to too many loci |" in line:
                data["multimapped_toomany_percent"] = float(
                    line.replace("% of reads mapped to too many loci |", "")
                    .strip()
                    .replace("%", "")
                )
    return data


outdir = sys.argv[1]
files = glob.glob(outdir + "/*.star.log")
data = []
samples = []
for file in files:
    _, sample = scp.ul.dir_base(file)
    samples.append(sample[: -len(".star")])
    d = get_numreads_percmapped(f"{outdir}/{sample}.log")
    data.append(d)
df = pd.DataFrame(data, index=samples)
df.to_csv(outdir + "/_starinfo.csv")



import math
import os
import sys
import time

import networkx as nx
import numpy as np
import pandas as pd

import scphylo as scp


def scelestial(df_input):
    """Solving using Scelestial.

    Fast and accurate single-cell lineage tree inference based on a Steiner tree
    approximation algorithm :cite:`Scelestial`.

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        Input genotype matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1), absence (0) and missing
        entires (3).

    Returns
    -------
    :class:`pandas.DataFrame`
        A conflict-free matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1) and absence (0).
    """
    executable = scp.ul.executable("scelestial", "Scelestial")

    scp.logg.info("running Scelestial")

    tmpdir = scp.ul.tmpdirsys(suffix=".scelestial")

    np.savetxt(
        f"{tmpdir.name}/scelestial.SC.T", df_input.values.T, delimiter="\t", fmt="%1.0f"
    )
    with open(f"{tmpdir.name}/scelestial.cellNames", "w") as fout:
        fout.write("\n".join(df_input.index))
    with open(f"{tmpdir.name}/scelestial.mutNames", "w") as fout:
        fout.write("\n".join(df_input.columns))

    _convert_input(
        f"{tmpdir.name}/scelestial.SC.T", f"{tmpdir.name}/scelestial.input", "/dev/null"
    )

    cmd = (
        f"{executable} < {tmpdir.name}/scelestial.input > "
        f"{tmpdir.name}/scelestial.tree_clone"
    )

    s_time = time.time()
    os.system(cmd)
    e_time = time.time()
    running_time = e_time - s_time

    _steiner_to_seq(
        f"{tmpdir.name}/scelestial.tree_clone",
        f"{tmpdir.name}/scelestial.impute",
    )

    _stein_to_clone_tree(
        f"{tmpdir.name}/scelestial.input",
        f"{tmpdir.name}/scelestial.tree_clone",
        f"{tmpdir.name}/scelestial.tree",
        f"{tmpdir.name}/scelestial.clone",
    )

    _clone_tree_to_mu_tree_imput(
        f"{tmpdir.name}/scelestial.tree",
        f"{tmpdir.name}/scelestial.clone",
        f"{tmpdir.name}/scelestial.SC.T",
        f"{tmpdir.name}/scelestial.mutNames",
        f"{tmpdir.name}/scelestial.cellNames",
        f"{tmpdir.name}/scelestial.output",
    )

    tree = nx.DiGraph(nx.nx_pydot.read_dot(f"{tmpdir.name}/scelestial.output"))
    tree = nx.relabel_nodes(tree, lambda x: int(x))
    tree.graph["splitter_mut"] = "\n"
    tree.graph["splitter_cell"] = "\n"
    tree.add_node(-1, label="root")
    tree.add_edge(-1, scp.ul.root_id(tree))
    for i, e in enumerate(tree.edges()):
        tree.edges[e]["label"] = f"mut{i}"
        tree.edges[e]["weight"] = 1
    for n in tree.nodes():
        if tree.nodes[n]["label"] == '""':
            tree.nodes[n]["label"] = "––"
    df = scp.ul.to_cfmatrix(tree)
    df2 = pd.DataFrame(
        np.identity(df.shape[0]),
        index=df.index,
        columns=[f"mut{x}" for x in range(df.shape[0], 2 * df.shape[0])],
        dtype=int,
    )
    df_output = df.merge(df2, right_index=True, left_index=True)

    tmpdir.cleanup()

    scp.ul.stat(df_output, df_output, 0, 0, running_time)

    return df_output


def _convert_input(sciteFile_path, imputeFile_path, bpFile_path):
    sciteFile = open(sciteFile_path)
    imputeFile = open(imputeFile_path, "w+")
    bpFile = open(bpFile_path, "w+")

    seq = []
    for line in sciteFile:
        x = line.strip().split()
        for i, v in enumerate(x):
            while i >= len(seq):
                seq.append([])
            seq[i].append(int(v))

    loc = len(seq[0])
    for i in range(loc):
        print(f"{i + 1}", end=" ", file=imputeFile)
        for j in range(len(seq)):
            s = seq[j][i]
            if s == 0:
                ss = "A/A"
            elif s == 1:
                ss = "C/C"
            elif s == 2:
                ss = "A/C"
            elif s == 3:
                ss = "./."
            else:
                print(f"OH! {s}")
                raise ValueError(f"OH! {s}")
            print(f"{ss}", end=" ", file=imputeFile)
        print(file=imputeFile)

    for i in range(loc):
        print(f"V{i + 1}", end="", file=bpFile)
        if i != loc - 1:
            print(",", end="", file=bpFile)
    print(file=bpFile)

    for s in seq:
        first = True
        for v in s:
            if v == 3:
                v = 2
            elif v == 2:
                v = 1
            if not first:
                print(",", end="", file=bpFile)
            first = False
            print(f"{v}", end="", file=bpFile)
        print(file=bpFile)

    sciteFile.close()
    imputeFile.close()
    bpFile.close()


def _stein_to_clone_tree(
    seqSciteFile_path, steinerFile_path, treeFile_path, cloneFile_path
):
    seqSciteFile = open(seqSciteFile_path)
    steinerFile = open(steinerFile_path)
    treeFile = open(treeFile_path, "w+")
    cloneFile = open(cloneFile_path, "w+")
    type(seqSciteFile)

    n = int(steinerFile.readline().strip())
    treeNodes = []
    cells = []
    maxAc = -1
    for _ in range(n):
        line = steinerFile.readline().strip()
        x = line.split()
        treeNodes.append(x[0])
        if x[1] == "1":
            cells.append(x[0])
        ac = sum(1 for s in x[2] if s == "A")
        if ac > maxAc:
            maxAc = ac
            maxAcLen = len(x[2])
            treeRootSeqIdx = int(x[0])

    seqIdxToExclude = -1
    if maxAcLen == maxAc and 5 < len(sys.argv) and sys.argv[5] == "-exclude-root":
        seqIdxToExclude = treeRootSeqIdx

    cells = set(cells)

    for t in treeNodes:
        print(t, end=" ", file=cloneFile)
        if t in cells and t != seqIdxToExclude and int(t) != seqIdxToExclude:
            print(int(t) + 1, end="", file=cloneFile)
        if int(t) == treeRootSeqIdx:
            treeRootCloneIdx = int(t)
        print(file=cloneFile)

    print(" ".join(treeNodes), file=treeFile)

    edges = {}

    m = int(steinerFile.readline().strip())
    for _ in range(m):
        line = steinerFile.readline().strip()
        x = line.split()
        if x[0] not in edges:
            edges[x[0]] = []
        if x[1] not in edges:
            edges[x[1]] = []
        edges[x[0]].append((x[1], x[2]))
        edges[x[1]].append((x[0], x[2]))

    mark = {}
    global dfsNumCounter
    dfsNumCounter = 0
    global dfsNum
    dfsNum = {}

    def dfs(v):
        global dfsNumCounter, dfsNum
        dfsNum[v] = dfsNumCounter
        dfsNumCounter += 1
        mark[v] = True
        for u, _ in edges[v]:
            if u not in mark:
                dfs(u)

    dfs(str(treeRootCloneIdx))

    for v, nei in edges.items():
        for u, w in nei:
            if v < u:
                x, y = u, v
                if dfsNum[v] > dfsNum[u]:
                    x, y = v, u
                print(f"{x}->{y} {w}", file=treeFile)

    seqSciteFile.close()
    steinerFile.close()
    treeFile.close()
    cloneFile.close()


def _clone_tree_to_mu_tree_imput(
    treeFileName,
    cloneFileName,
    seqFileName,
    mutationInfoFileName,
    cellNamesFileName,
    outputFileName,
    margeClones=False,
    markMutations=False,
    compress=False,
    markMutationsSeparated=False,
):
    from graphviz import Digraph
    from scipy import stats

    def loadTree(treeFileName):
        treeFile = open(treeFileName)
        line = treeFile.readline()
        vertices = line.strip().split()
        edges = []
        treeParent = {}
        treeChildren = {}
        for line in treeFile:
            (e, w) = line.strip().split(" ")
            (v, u) = e.split("->")
            edges.append((v, u, float(w)))

            treeParent[v] = u
            if u not in treeChildren:
                treeChildren[u] = []
            treeChildren[u].append((v, float(w)))

        treeRoot = list(treeParent.keys())[0]
        while treeRoot in treeParent:
            treeRoot = treeParent[treeRoot]

        treeFile.close()
        return vertices, edges, treeParent, treeChildren, treeRoot

    def loadClones(cloneFileName):
        cloneFile = open(cloneFileName)
        treeNodeCells = {}

        for line in cloneFile:
            x = line.strip().split()
            treeNodeCells[x[0]] = x[1:]

        cloneFile.close()
        return treeNodeCells

    def compressedTree(treeChildren, treeNodeCells, treeRoot):
        compressedTreeChildren = {}

        def dfs(v):
            ret = (None, None)
            children = []
            if v in treeChildren:
                for u, cw in treeChildren[v]:
                    c, w = dfs(u)
                    if c is not None:
                        w += cw
                        children.append((c, w))
                        ret = (c, w)

            if (
                (v in treeNodeCells and len(treeNodeCells[v]) > 0)
                or len(children) > 1
                or v == treeRoot
            ):
                ret = (v, 0)

            if ret == (v, 0):
                compressedTreeChildren[v] = []
                for c, cw in children:
                    compressedTreeChildren[v].append((c, cw))

            return ret

        dfs(treeRoot)
        return compressedTreeChildren

    def loadSequenceFile(seqFileName):
        sequences = []
        seqFile = open(seqFileName)
        for line in seqFile:
            for num, val in enumerate(line.strip().split()):
                while num >= len(sequences):
                    sequences.append([])
                sequences[num].append(int(val))
        seqFile.close()
        return sequences

    def loadMutationInfoFile(mutationInfoFileName):
        mutationInfoFile = open(mutationInfoFileName)
        mutationInfo = []
        for line in mutationInfoFile:
            mutationInfo.append({id: line.strip()})
        mutationInfoFile.close()
        return mutationInfo

    def writeGraph(
        outputFileName, treeNodeCells, nodes, edges, treeNodeDescColor, treeEdgeLabel
    ):
        dot = Digraph(format="pdf")
        dot.graph_attr["rankdir"] = "LR"
        for treeNode in nodes:
            cells = []
            if treeNode in treeNodeCells:
                cells = treeNodeCells[treeNode]
            prop = treeNodeDescColor(treeNode, cells)
            if len(prop) == 2:
                desc, col, fontcol, fillcol = prop[0], prop[1], "black", "none"
            else:
                desc, col, fontcol, fillcol = prop[0], prop[1], prop[2], prop[3]
            dot.node(
                treeNode,
                desc,
                color=col,
                fillcolor=fillcol,
                style="filled",
                fontcolor=fontcol,
                gradientangle="0",
                penwidth="4",
                shape="circle",
                margin="0",
            )

        for v, u, w in edges:
            tup = treeEdgeLabel(v, u, w)
            if isinstance(tup, tuple):
                label, edgecol = tup[0], tup[1]
            else:
                label, edgecol = tup, "black"
            dot.edge(u, v, weight=str(w), label=label, color=edgecol)

        dot.render(outputFileName)

    def loadFileRows(fileName):
        f = open(fileName)
        r = []
        for line in f:
            r.append(line.strip())
        f.close()
        return r

    def loadCellNames(cellNamesFileName):
        return loadFileRows(cellNamesFileName)

    BLUE = "#69c5f0"
    BLACK = "black"

    _, edges, _, treeChildren, treeRoot = loadTree(treeFileName)
    treeNodeCells = loadClones(cloneFileName)

    sequences = loadSequenceFile(seqFileName)
    mutationInfo = loadMutationInfoFile(mutationInfoFileName)
    cellNames = loadCellNames(cellNamesFileName)

    if margeClones:

        def sameColon(v, u, w):
            m = len(sequences[0])
            pv = stats.binom_test(w, m, 0.2, alternative="less")
            print(f"same colon test: {w}/{m} = {pv}")
            return pv < 0.01

        mergeParent = {v: v for v in treeChildren.keys()}
        toBeMerged = {}
        for v, childrenDistList in treeChildren.items():
            for c, w in childrenDistList:
                if sameColon(v, c, w):
                    mergeParent[c] = v
                    toBeMerged[c] = True

        newTreeChildren = {v: [] for v in treeChildren.keys()}
        newTreeNodeCells = {v: [] for v in treeNodeCells.keys()}

        def dfs(v, firstKeptNode, wToHere):
            newTreeNodeCells[firstKeptNode] += treeNodeCells[v]
            if v in treeChildren:
                for u, w in treeChildren[v]:
                    if u not in toBeMerged:
                        newTreeChildren[firstKeptNode].append((u, wToHere + w))
                        dfs(u, u, 0)
                    else:
                        dfs(u, firstKeptNode, wToHere + w)

        dfs(treeRoot, treeRoot, 0)

        for v, _ in treeChildren.items():
            while mergeParent[mergeParent[v]] != mergeParent[v]:
                mergeParent[v] = mergeParent[mergeParent[v]]

        treeNodeCells = newTreeNodeCells
        treeChildren = newTreeChildren

    treeNodeMutations = {}

    def fillTreeNodeMutations(mutIndex=None):
        def dfs(v):
            myCellsStar = []
            if v in treeNodeCells:
                myCellsStar += [int(u) - 1 for u in treeNodeCells[v]]
            if v in treeChildren:
                for u, _ in treeChildren[v]:
                    myCellsStar = dfs(u) + myCellsStar

            nodeMutations[v] = []
            for i, _ in enumerate(mutationInfo):
                subTreeNormal, subTreeMutated = 0, 0
                for c in myCellsStar:
                    if sequences[c][i] == 0:
                        subTreeNormal += 1
                    if sequences[c][i] == 1:
                        subTreeMutated += 1

                if subTreeMutated > 0:
                    ignoremut = False
                    if v in treeChildren:
                        for u, _ in treeChildren[v]:
                            smi = [
                                submut for ii, submut, _ in nodeMutations[u] if i == ii
                            ]
                            if len(smi) == 1 and smi[0] == subTreeMutated:
                                ignoremut = True
                    if not ignoremut:
                        nodeMutations[v].append(
                            (
                                i,
                                subTreeMutated,
                                str(subTreeNormal) + "," + str(len(myCellsStar)),
                            )
                        )
            return myCellsStar

        nodeMutations = {}
        dfs(treeRoot)
        for v, muts in nodeMutations.items():
            treeNodeMutations[v] = [
                mutationInfo[i][id] + "/" + str(mut) + "," + str(desc)
                for i, mut, desc in muts
                if mutIndex is None or i == mutIndex
            ]

    if markMutations:
        fillTreeNodeMutations()

    nodes = list(treeNodeCells.keys())
    if compress:
        compressedTreeChildren = compressedTree(treeChildren, treeNodeCells, treeRoot)
        edges = []
        for par, cwList in compressedTreeChildren.items():
            for c, w in cwList:
                edges.append((c, par, w))
        nodes = list(compressedTreeChildren.keys())

    def treeNodeDescColor(treeNode, cells):
        desc = ", ".join([cellNames[int(c) - 1] for c in cells])
        if treeNodeMutations is not None and treeNode in treeNodeMutations:
            mutations = sorted(set(treeNodeMutations[treeNode]))
            seqmut = math.sqrt(len(mutations) / 7)
            if len(mutations) > 0:
                desc += " ["
                lastLineLen = 0
                for _, mut in enumerate(mutations):
                    desc += mut + " "
                    lastLineLen += 1
                    if lastLineLen >= seqmut:
                        desc += "\n"
                        lastLineLen = 0
                desc += "]"
        return desc, BLUE, BLACK, BLUE

    def treeEdgeLabel(v, u, w):
        return "", "#A2A2A2"

    if markMutationsSeparated:
        for i, _ in enumerate(mutationInfo):
            fillTreeNodeMutations(i)
            writeGraph(
                outputFileName + "-" + str(i),
                treeNodeCells,
                nodes,
                edges,
                treeNodeDescColor,
                treeEdgeLabel,
            )
    else:
        writeGraph(
            outputFileName,
            treeNodeCells,
            nodes,
            edges,
            treeNodeDescColor,
            treeEdgeLabel,
        )


def _steiner_to_seq(steinerFile_path, imputeFile_path):
    steinerFile = open(steinerFile_path)
    imputeFile = open(imputeFile_path, "w+")
    n = int(steinerFile.readline().strip())
    treeNodes = []
    cells = []
    maxAc = -1
    seq = []
    for _ in range(n):
        line = steinerFile.readline().strip()
        x = line.split()
        treeNodes.append(x[0])
        if x[1] == "1":
            cells.append(x[0])
        ac = sum(1 for s in x[2] if s == "A")
        if ac >= maxAc:
            maxAc = ac

        if x[1] == "1":
            if sum(1 for s in x[3] if s not in {"A", "C"}) > 0:
                print(
                    f"imputed sequence contains invalid char {x[3]}",
                    file=sys.stderr,
                )
                raise Exception("Invalid imputation")
            seq.append(x[3])

    if len(seq) == 0:
        exit()
    for j in range(len(seq[0])):
        print(
            " ".join(["0" if s[j] == "A" else "1" for s in seq]),
            file=imputeFile,
        )
    steinerFile.close()
    imputeFile.close()



"""Score Module."""

from scphylo.tl.score._others import caset, disc, mp3, rf
from scphylo.tl.score._ours import ad, cc, dl, gs, mltd, tpted

__all__ = (caset, disc, mp3, rf, ad, cc, dl, gs, mltd, tpted)



import scphylo as scp


def import_gurobi():
    try:
        import gurobipy as gp

        return gp, False
    except ImportError:
        scp.logg.warn(
            "Unable to import `gurobipy`!",
            "Make sure `Gurobi` is already installed in your system.",
            "Then install `gurobipy`.",
        )
        return None, True


def import_mpi4py():
    try:
        import mpi4py

        return mpi4py, False
    except ImportError:
        scp.logg.warn(
            "Unable to import `mpi4py`!",
            "Make sure `mpi/mpich-x86_64` is already installed in your system.",
            "Then install `mpi4py` using `pip`.",
        )
        return None, True


def import_rpy2(name="base", how=""):
    try:
        import logging

        from rpy2.rinterface_lib.callbacks import logger as rpy2_logger
        from rpy2.robjects import r
        from rpy2.robjects.packages import PackageNotInstalledError, importr

        type(r)
        rpy2_logger.setLevel(logging.ERROR)

    except ImportError:
        scp.logg.warn(
            "Unable to import `rpy2`, install it first as `pip install rpy2` version"
            " `>=3.3.0`."
        )
        return None, True

    try:
        _r_lib = importr(name)
        return _r_lib, False
    except PackageNotInstalledError:
        scp.logg.warn(f"Install R library `{name!r}` first.\n{how}")
        return None, True


def import_graphviz():
    try:
        import pygraphviz

        return pygraphviz, False
    except ImportError:
        scp.logg.warn(
            "Unable to import `pygraphviz`!",
            "Make sure `Graphviz` is already installed in your system.",
            "Then install `pygraphviz` using `pip`.",
        )
        return None, True


def import_graph_tool():
    try:
        import graph_tool

        return graph_tool, False
    except ImportError:
        scp.logg.warn(
            "Unable to import `graph_tool`!",
            "Make sure `graph_tool` is installed in your system.",
        )
        return None, True



"""Utils Module."""

from scphylo.ul._caller import get_samples_df, is_paired
from scphylo.ul._hclustering import (
    dist_cosine_ignore_na,
    dist_dendro,
    dist_l1_ignore_na,
    hclustering,
)
from scphylo.ul._packages import (
    import_graph_tool,
    import_graphviz,
    import_gurobi,
    import_mpi4py,
    import_rpy2,
)
from scphylo.ul._trees import (
    cells_rooted_at,
    is_leaf,
    muts_rooted_at,
    partition_cells,
    root_id,
    to_cfmatrix,
    to_mtree,
    to_tree,
)
from scphylo.ul._utils import (
    calc_nll_matrix,
    cleanup,
    count_flips,
    dir_base,
    dirbase,
    executable,
    get_file,
    infer_rates,
    is_conflict_free,
    is_conflict_free_gusfield,
    log_flip,
    log_input,
    log_output,
    mkdir,
    parse_log_file,
    parse_params_file,
    parse_score_file,
    remove,
    split_mut,
    stat,
    timeit,
    tmpdir,
    tmpdirsys,
    tmpfile,
    tqdm_joblib,
    with_timeout,
)

__all__ = (
    dist_cosine_ignore_na,
    dist_dendro,
    dist_l1_ignore_na,
    hclustering,
    import_graph_tool,
    import_graphviz,
    import_gurobi,
    import_mpi4py,
    import_rpy2,
    cells_rooted_at,
    muts_rooted_at,
    partition_cells,
    root_id,
    to_cfmatrix,
    to_mtree,
    to_tree,
    calc_nll_matrix,
    cleanup,
    dir_base,
    dirbase,
    get_file,
    parse_params_file,
    parse_log_file,
    parse_score_file,
    infer_rates,
    is_conflict_free,
    is_conflict_free_gusfield,
    log_flip,
    log_input,
    log_output,
    mkdir,
    remove,
    stat,
    timeit,
    tmpdir,
    tmpdirsys,
    tmpfile,
    tqdm_joblib,
    with_timeout,
    is_leaf,
    executable,
    split_mut,
    count_flips,
    is_paired,
    get_samples_df,
)



"""Huntress Module."""

import pandas as pd

import scphylo as scp
from scphylo.tl.solver.huntress._huntress import Reconstruct


def huntress(df_input, alpha, beta, n_threads=1):
    """Solving using HUNTRESS.

    HUNTRESS: Provably fast intratumor heterogeneity inference from single-cell
    sequencing data

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        Input genotype matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1), absence (0) and missing
        entires (3).
    alpha : :obj:`float`
        False positive error rate.
    beta : :obj:`float`
        False negative error rate.
    n_threads : :obj:`int`, optional
        Number of threads, by default 1

    Returns
    -------
    :class:`pandas.DataFrame`
        A conflict-free matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1) and absence (0).
    """
    scp.logg.info(
        f"running HUNTRESS with alpha={alpha}, beta={beta}, n_threads={n_threads}"
    )

    running_time = 0
    if alpha == 0:
        matrix_out, running_time = Reconstruct(
            df_input,
            Algchoice="FN",
            n_proc=n_threads,
        )
    else:
        fn_conorm = 0.1
        fp_conorm = fn_conorm * alpha / beta
        matrix_out, running_time = Reconstruct(
            df_input,
            n_proc=n_threads,
            fnfp=51,
            post_fn=fn_conorm,
            post_fp=fp_conorm,
        )

    df_output = pd.DataFrame(matrix_out)
    df_output.columns = df_input.columns
    df_output.index = df_input.index
    df_output.index.name = "cellIDxmutID"

    scp.ul.stat(df_input, df_output, alpha, beta, running_time)

    return df_output



import click
from joblib import Parallel, delayed
from tqdm import tqdm

import scphylo as scp


def function(patient, tqdm_pos, kind):
    """[summary].

    Parameters
    ----------
    patient : [type]
        [description]
    tqdm_pos : [type]
        [description]
    """
    # for _ in tqdm(
    #     range(10), ascii=True, ncols=60, desc=f"", leave=False, position=pos + 1
    # ):
    #     sleep(1)
    scp.io.build_crc(patient=patient, tqdm_pos=tqdm_pos, kind=kind)


@click.command(short_help="Building the matrices of CRCs.")
def build_crcs():
    """[summary].

    Returns
    -------
    [type]
        [description]
    """
    patients = ["CRC01", "CRC02", "CRC04", "CRC09", "CRC10", "CRC11"]
    kinds = ["FPKM", "FPKM", "TPM", "TPM", "TPM", "TPM"]

    # only met
    # "CRC12", "CRC13", "CRC14", "CRC15"

    # only expr
    # "CRC03", "CRC06"

    number_of_samples = len(patients)
    number_of_threads = len(patients)
    with scp.ul.tqdm_joblib(
        tqdm(
            ascii=True,
            ncols=75,
            desc="STATUS",
            total=number_of_samples,
            position=0,
        )
    ) as _:
        Parallel(n_jobs=number_of_threads)(
            delayed(function)(patients[i], i + 1, kinds[i])
            for i in range(number_of_samples)
        )

    # function("CRC01", 1, "FPKM")

    return None



# Copyright Ian Smith 2002-2018
# Above only included as I am told that without it, others may claim copyright and stop
# me from editting my own copies of my own code! No idea how true this is.
# Beyond that, I am happy for anyone to make any edits they like to this code.
# For example, raising ValueError in pmf, cdf and sf functions should perhaps only be
# done when the distribution objects are created

# This code is basically translated from the VBA code available
# at https://iandjmsmith.wordpress.com

import math

NonIntegralValuesAllowed_Others = False

OneOverSqrTwoPi = 0.39894228040143267793994605993438
cfSmall = 0.00000000000001
cfVSmall = 0.000000000000001
scalefactor = 1.1579208923731619542357098500869e77  # 2**256
scalefactor2 = 8.6361685550944446253863518628004e-78  # 2**-256
max_discrete = 9007199254740991.0
minLog1Value = -0.79149064

HalfMinusEulers_const = -0.07721566490153286060651209011
Onep25Minusln2Minuseulers_const = -0.020362845461478170023744211558177
coeffs0Minusp25 = 0.07246703342411321823620758332301

# For logfbit functions
# Stieltjes' continued fraction
cf_0 = 1.0 / 12.0
cf_1 = 1.0 / 30.0
cf_2 = 53.0 / 210.0
cf_3 = 195.0 / 371.0
cf_4 = 22999.0 / 22737.0
cf_5 = 29944523.0 / 19733142.0
cf_6 = 109535241009.0 / 48264275462.0
cf_7 = 3.0099173832593981700731407342077
cf_8 = 4.026887192343901226168879531814
cf_9 = 5.0027680807540300516885024122767
cf_10 = 6.2839113708157821800726631549524
cf_11 = 7.4959191223840339297523547082674
cf_12 = 9.0406602343677266995311393604326
cf_13 = 10.489303654509482277188371304593
cf_14 = 12.297193610386205863989437140092
cf_15 = 13.982876953992430188259760651279
cf_16 = 16.053551416704935469715616365007
cf_17 = 17.976607399870277592569472307671
cf_18 = 20.309762027441653743805414720495
cf_19 = 22.470471639933132495517941571508
cf_20 = 25.065846548945972029163400322506
cf_21 = 27.464451825029133609175558982646
cf_22 = 30.321821231673047126882599306406
cf_23 = 32.958533929972987219994066451412
cf_24 = 36.077698931299242645153320900855
cf_25 = 38.952706682311555734544390410481
cf_26 = 42.333490043576957211381853948856
cf_27 = 45.446960850061621014424175737541
cf_28 = 49.089203129012597708164883350275
cf_29 = 52.441288751415337312569856046996
cf_30 = 56.344845345341843538420365947476
cf_31 = 59.935683907165858207852583492752
cf_32 = 64.100422755920354527906611892238
cf_33 = 67.930140788018221186367702745199
cf_34 = 72.355940555211701969680052963236
cf_35 = 76.424654626829689752585090422288
cf_36 = 81.111403237247965484814230985683
cf_37 = 85.419221276410972614585638717349
cf_38 = 90.366814723864108595513574581683
cf_39 = 94.913837100009887953076231291987
cf_40 = 100.12217846392919748899074683447

lfbArray = [
    cf_0,
    cf_1,
    cf_2,
    cf_3,
    cf_4,
    cf_5,
    cf_6,
    cf_7,
    cf_8,
    cf_9,
    cf_10,
    cf_11,
    cf_12,
    cf_13,
    cf_14,
    cf_15,
    cf_16,
    cf_17,
    cf_18,
    cf_19,
    cf_20,
    cf_21,
    cf_22,
    cf_23,
    cf_24,
    cf_25,
    cf_26,
    cf_27,
    cf_28,
    cf_29,
]
coeffs = [
    0.32246703342411321824,
    6.7352301053198095133e-02,
    2.0580808427784547879e-02,
    7.3855510286739852663e-03,
    2.8905103307415232858e-03,
    1.1927539117032609771e-03,
    5.0966952474304242234e-04,
    2.2315475845357937976e-04,
    9.9457512781808533715e-05,
    4.4926236738133141700e-05,
    2.0507212775670691553e-05,
    9.4394882752683959040e-06,
    4.3748667899074878042e-06,
    2.0392157538013662368e-06,
    9.5514121304074198329e-07,
    4.4924691987645660433e-07,
    2.1207184805554665869e-07,
    1.0043224823968099609e-07,
    4.7698101693639805658e-08,
    2.2711094608943164910e-08,
    1.0838659214896954091e-08,
    5.1834750419700466551e-09,
    2.4836745438024783172e-09,
    1.1921401405860912074e-09,
    5.7313672416788620133e-10,
    2.7595228851242331452e-10,
    1.3304764374244489481e-10,
    6.4229645638381000221e-11,
    3.1044247747322272762e-11,
    1.5021384080754142171e-11,
    7.2759744802390796625e-12,
    3.5277424765759150836e-12,
    1.7119917905596179086e-12,
    8.3153858414202848198e-13,
    4.0422005252894400655e-13,
    1.9664756310966164904e-13,
    9.5736303878385557638e-14,
    4.6640760264283742246e-14,
    2.2737369600659723206e-14,
    1.1091399470834522017e-14,
    5.4136591567253631315e-15,
    2.6438800178609949985e-15,
    1.2918959062789967293811764562316e-15,
    6.3159355041984485676779394847024e-16,
    1.421085482803160676983430580383e-14,
]
coeffs2 = [
    9.96703342411321823620758332301e-3,
    4.85230105319809513323333333333e-3,
    2.35164176111788121233425746862e-3,
    1.1355510286739852662730972914e-3,
    5.4676033074152328575298829848e-4,
    2.6269438789373716759012616902e-4,
    1.2601997117161385090708338501e-4,
    6.03943417869127130948e-5,
    2.89279988929196448257e-5,
    1.38537935563149598820e-5,
    6.63558635521614609862e-6,
    3.17947224962737026300e-6,
    1.52432377823166362836e-6,
    7.31319548444223379639e-7,
    3.51147479316783649952e-7,
    1.68754473874618369035e-7,
    8.11753732546888155551e-8,
    3.90847775936649142159e-8,
    1.88369052760822398681e-8,
    9.08717580313959348175e-9,
    4.38794008336117216467e-9,
    2.12078578473653627963e-9,
    1.02595225309999020577e-9,
    4.96752618206533915776e-10,
    2.40726205228207715756e-10,
    1.16751848407213311847e-10,
    5.66693373586358102002e-11,
    2.75272781658498271913e-11,
    1.33812334011666457419e-11,
    6.50929603319331702812e-12,
    3.16857905287746826547e-12,
    1.54339039998043529180e-12,
    7.52239805801019839357e-13,
    3.66855576849641617566e-13,
    1.79011840661361776213e-13,
    8.73989502840846835258e-14,
    4.26932273880725309600e-14,
    2.08656067727432658676e-14,
    1.02026669800712891581e-14,
    4.99113012967463749398e-15,
    2.44274876330334144841e-15,
    1.19604099925791156327e-15,
    5.85859784064943690566e-16,
    2.87087981566462948467e-16,
    1.40735520163755175636e-16,
]

logfbit0p5 = 0.054814121051917653896138702348386  # logfbit(0.5)
lfb_0 = 8.1061466795327258219670263594382e-02  # logfbit(0.0)
lfb_1 = 4.1340695955409294093822081407118e-02  # logfbit(1.0)
lfb_2 = 2.7677925684998339148789292746245e-02  # logfbit(2.0)
lfb_3 = 2.0790672103765093111522771767849e-02  # logfbit(3.0)
lfb_4 = 1.6644691189821192163194865373593e-02  # logfbit(4.0)
lfb_5 = 1.3876128823070747998745727023763e-02  # logfbit(5.0)

# For logfbit functions
# Stirling's series for ln(Gamma(x)), A046968/A046969
lfbc1 = 1.0 / 12.0
lfbc2 = 1.0 / 30.0  # lfbc2 on are Sloane's ratio times 12
lfbc3 = 1.0 / 105.0
lfbc4 = 1.0 / 140.0
lfbc5 = 1.0 / 99.0
lfbc6 = 691.0 / 30030.0
lfbc7 = 1.0 / 13.0
lfbc8 = 0.35068485511628418514  # Chosen to make logfbit(6) & logfbit(7) correct
lfbc9 = 1.6769380337122674863  # Chosen to make logfbit(6) & logfbit(7) correct

# Global variables used by functions hypergeometric & CBNB0
ha1, hprob, hswap = 0.0, 0.0, False


def AlterForIntegralChecks_Others(value):
    # If non integral values are allowed then returns math.floor(value) otherwise only
    # allows value if it equals math.floor(value)
    if NonIntegralValuesAllowed_Others:
        return math.floor(value)
    elif value != math.floor(value):
        raise ValueError
    else:
        return value


def Generalabminuscd(a, b, c, d):
    if (a == 0.0) or (b == 0.0) or (c == 0.0) or (d == 0.0):
        return a * b - c * d
    amr, ae = math.frexp(a)
    bmr, be = math.frexp(b)
    cmr, ce = math.frexp(c)
    dmr, de = math.frexp(d)
    am = int(amr * (2**53))
    bm = int(bmr * (2**53))
    cm = int(cmr * (2**53))
    dm = int(dmr * (2**53))
    abe = ae + be
    cde = ce + de
    abm = am * bm
    cdm = cm * dm
    scale = abe - cde
    if scale >= 0:
        if scale > 106:
            scale = 0
            cdm = 1
        # return math.ldexp(abm * (2**scale) - cdm, abe - scale - 106)
        return float(abm * (2**scale) - cdm) * (2.0 ** (abe - scale - 106))
    else:
        scale = -scale
        if scale > 106:
            scale = 0
            abm = 1
        # return math.ldexp(abm - cdm * (2**scale), cde - scale - 106)
        return float(abm - cdm * (2**scale)) * (2.0 ** (cde - scale - 106))


def logfbit2dif(x):
    # Calculation of logfbit2(x)-logfbit2(1+x).
    return 0.5 * (((x + 1.0) * (x + 2.0)) ** -2)


def logfbit2(x):
    # Second derivative of error part of Stirling's formula where
    # log(x!) = log(sqrt(twopi))+(x+0.5)*log(x+1)-(x+1)+logfbit(x).
    if x >= 10000000000.0:
        return 2.0 * lfbc1 * ((x + 1.0) ** -3)
    elif x >= 7.0:
        x1 = x + 1.0
        x2 = 1.0 / (x1 * x1)
        x3 = x2 * (240.0 * lfbc8 - x2 * 306.0 * lfbc9)
        x3 = x2 * (132.0 * lfbc6 - x2 * (182.0 * lfbc7 - x3))
        x3 = x2 * (56.0 * lfbc4 - x2 * (90.0 * lfbc5 - x3))
        x3 = x2 * (12.0 * lfbc2 - x2 * (30.0 * lfbc3 - x3))
        return lfbc1 * (2.0 - x3) * x2 / x1
    elif x > -1.0:
        x1 = x
        x2 = 0.0
        while x1 < 7.0:
            x2 = x2 + logfbit2dif(x1)
            x1 = x1 + 1.0
        return x2 + logfbit2(x1)
    else:
        raise ValueError


def logfbit4dif(x):
    # Calculation of logfbit4(x)-logfbit4(1+x).
    return (10.0 * x * (x + 3.0) + 23.0) * (((x + 1.0) * (x + 2.0)) ** -4)


def logfbit4(x):
    # Fourth derivative of error part of Stirling's formula where
    # log(x!) = log(sqrt(twopi))+(x+0.5)*log(x+1)-(x+1)+logfbit(x).
    if x >= 10000000000.0:
        return -0.5 * ((x + 1.0) ** -4)
    elif x >= 7.0:
        x1 = x + 1.0
        x2 = 1.0 / (x1 * x1)
        x3 = x2 * (73440.0 * lfbc8 - x2 * 116280.0 * lfbc9)
        x3 = x2 * (24024.0 * lfbc6 - x2 * (43680.0 * lfbc7 - x3))
        x3 = x2 * (5040.0 * lfbc4 - x2 * (11880.0 * lfbc5 - x3))
        x3 = x2 * (360.0 * lfbc2 - x2 * (1680.0 * lfbc3 - x3))
        return lfbc1 * (24.0 - x3) * x2 * x2 / x1
    elif x > -1.0:
        x1 = x
        x2 = 0.0
        while x1 < 7.0:
            x2 = x2 + logfbit4dif(x1)
            x1 = x1 + 1.0

        return x2 + logfbit4(x1)
    else:
        raise ValueError


def log0(x):
    # Accurate and quicker calculation of log(1+x), particularly for small x.
    # Code from Wolfgang Ehrhardt.
    if x > 4.0:
        return math.log(1.0 + x)
    else:
        y = 1.0 + x
        if y == 1.0:
            return x
        else:
            return math.log(y) + (x - (y - 1.0)) / y


def logcf(x, i, d):
    # Continued fraction for calculation of
    # 1/i + x/(i+d) + x*x/(i+2*d) + x*x*x/(i+3d) + ...
    c1 = 2.0 * d
    c2 = i + d
    c4 = c2 + d
    a1 = c2
    b1 = i * (c2 - i * x)
    b2 = d * d * x
    a2 = c4 * c2 - b2
    b2 = c4 * b1 - i * b2

    while abs(a2 * b1 - a1 * b2) > abs(cfVSmall * b1 * a2):

        c3 = c2 * c2 * x
        c2 = c2 + d
        c4 = c4 + d
        a1 = c4 * a2 - c3 * a1
        b1 = c4 * b2 - c3 * b1

        c3 = c1 * c1 * x
        c1 = c1 + d
        c4 = c4 + d
        a2 = c4 * a1 - c3 * a2
        b2 = c4 * b1 - c3 * b2
        if b2 > scalefactor:
            a1 = a1 * scalefactor2
            b1 = b1 * scalefactor2
            a2 = a2 * scalefactor2
            b2 = b2 * scalefactor2
        elif b2 < scalefactor2:
            a1 = a1 * scalefactor
            b1 = b1 * scalefactor
            a2 = a2 * scalefactor
            b2 = b2 * scalefactor
    return a2 / b2


def log1(x):
    # Accurate calculation of log(1+x)-x, particularly for small x.
    if abs(x) < 0.01:
        term = x / (2.0 + x)
        y = term * term
        return term * (
            (((2.0 / 9.0 * y + 2.0 / 7.0) * y + 0.4) * y + 2.0 / 3.0) * y - x
        )
    elif x < minLog1Value or x > 1.0:
        return math.log(1.0 + x) - x
    else:
        term = x / (2.0 + x)
        y = term * term
        return term * (2.0 * y * logcf(y, 3.0, 2.0) - x)


def logfbitdif(x):
    # Calculation of logfbit(x)-logfbit(1+x). x must be > -1.
    if x < -0.65:
        return (x + 1.5) * log0(1.0 / (x + 1.0)) - 1.0
    else:
        y2 = (2.0 * x + 3.0) ** -2
        return y2 * logcf(y2, 3.0, 2.0)


def logfbita(x):
    # Error part of Stirling's formula where math.log(x!) =
    # math.log(math.sqrtt(twopi))+(x+0.5)*math.log(x+1)-(x+1)+logfbita(x).
    # Are we ever concerned about the relative error involved in this function?
    # I don't think so.
    if x >= 100000000.0:
        return lfbc1 / (x + 1.0)
    elif x >= 6.0:  # Abramowitz & Stegun's series 6.1.41
        x1 = x + 1.0
        x2 = 1.0 / (x1 * x1)
        x3 = x2 * (lfbc6 - x2 * (lfbc7 - x2 * (lfbc8 - x2 * lfbc9)))
        x3 = x2 * (lfbc4 - x2 * (lfbc5 - x3))
        x3 = x2 * (lfbc2 - x2 * (lfbc3 - x3))
        return lfbc1 * (1.0 - x3) / x1
    elif x == 0.0:
        return lfb_0
    elif x == 1.0:
        return lfb_1
    elif x == 2.0:
        return lfb_2
    elif x == 3.0:
        return lfb_3
    elif x == 4.0:
        return lfb_4
    elif x == 5.0:
        return lfb_5
    elif x > -1.0:
        x1 = x
        x2 = 0.0
        while x1 < 6.0:
            x2 = x2 + logfbitdif(x1)
            x1 = x1 + 1.0
        return x2 + logfbit(x1)
    else:
        raise ValueError


def logfbit(x):
    # Calculates log of x factorial - log(sqrt(2*pi)) +(x+1) -(x+0.5)*log(x+1)
    # using the error part of Stirling's formula
    # (see Abramowitz & Stegun# s series 6.1.41)
    # and Stieltjes' continued fraction for the gamma function.
    # For x < 1.5, uses expansion of log(x!) and log((x+1)!)
    # from Abramowitz & Stegun# s series 6.1.33
    # We are primarily concerned about the absolute error in this function.
    # Due to cancellation errors in calculating 1+x as x tends to -1,
    # the function loses accuracy and should not be used!
    if x >= 6.0:
        x1 = x + 1.0
        if x >= 1000.0:
            if x >= 100000000.0:
                x3 = 0.0
            else:
                x2 = 1.0 / (x1 * x1)
                x3 = x2 * (lfbc2 - x2 * lfbc3)

        else:
            x2 = 1.0 / (x1 * x1)
            if x >= 40.0:
                x3 = 0.0
            elif x >= 15.0:
                x3 = x2 * (lfbc6 - x2 * lfbc7)
            else:
                x3 = x2 * (lfbc6 - x2 * (lfbc7 - x2 * (lfbc8 - x2 * lfbc9)))

            x3 = x2 * (lfbc4 - x2 * (lfbc5 - x3))
            x3 = x2 * (lfbc2 - x2 * (lfbc3 - x3))

        return lfbc1 * (1.0 - x3) / x1
        # return (1.0 - x3) / (12.0 * x1)
    elif x == 0.0:
        return lfb_0
    elif x == 1.0:
        return lfb_1
    elif x == 2.0:
        return lfb_2
    elif x == 3.0:
        return lfb_3
    elif x == 4.0:
        return lfb_4
    elif x == 5.0:
        return lfb_5
    elif x > 1.5:
        x1 = x + 1.0
        if x >= 2.5:
            x2 = 40.5 / (x1 + math.sqrt(x1 * x1 + 81.0))
        else:
            x2 = 112.5 / (x1 + math.sqrt(x1 * x1 + 225.0))
            x2 = cf_27 / (x1 + cf_28 / (x1 + cf_29 / (x1 + x2)))
            x2 = cf_24 / (x1 + cf_25 / (x1 + cf_26 / (x1 + x2)))
            x2 = cf_21 / (x1 + cf_22 / (x1 + cf_23 / (x1 + x2)))
            x2 = cf_18 / (x1 + cf_19 / (x1 + cf_20 / (x1 + x2)))

        x2 = cf_15 / (x1 + cf_16 / (x1 + cf_17 / (x1 + x2)))
        x2 = cf_12 / (x1 + cf_13 / (x1 + cf_14 / (x1 + x2)))
        x2 = cf_9 / (x1 + cf_10 / (x1 + cf_11 / (x1 + x2)))
        x2 = cf_6 / (x1 + cf_7 / (x1 + cf_8 / (x1 + x2)))
        x2 = cf_3 / (x1 + cf_4 / (x1 + cf_5 / (x1 + x2)))
        return 1.0 / (12.0 * (x1 + cf_1 / (x1 + cf_2 / (x1 + x2))))
    elif x == 0.5:
        return logfbit0p5  # 5.481412105191765389613870234839e-02
    elif x == -0.5:
        return 0.15342640972002734529138393927091  # 0.15342640972002734529138393927091
    elif x >= -0.65:
        if x <= 0.0:
            i = len(coeffs) - 1
            lgam = coeffs[i] * logcf(-x / 2.0, i + 2.0, 1.0)
            # print(i, coeffs[i],lgam)
            for i in range(i - 1, 0, -1):
                lgam = coeffs[i] - x * lgam
                # print(i,lgam)
            return (
                (
                    coeffs0Minusp25
                    - (x * (1.0 / 3.0 - (x + 1.5) * logcf(-x, 4.0, 1.0)) + lgam) * x
                )
                * x
                + HalfMinusEulers_const
            ) * x + lfb_0
        elif x <= 1.56:
            x = x - 1.0
            i = len(coeffs2) + 2
            lgam = ((x + 2.5) * logcf(-x / 2.0, i, 1.0) - (2.0 / (i - 1.0))) * (
                2.0**-i
            ) + (3.0**-i) * logcf(-x / 3.0, i, 1.0)
            # print(i,coeffs2[i-3],lgam)
            for i in range(i - 3, -1, -1):
                lgam = coeffs2[i] - x * lgam
                # print(i,lgam)
            return (x * lgam + Onep25Minusln2Minuseulers_const) * x + lfb_1
        elif x <= 2.5:
            x = x - 2.0
            i = len(coeffs) - 1
            lgam = coeffs[i] * logcf(-x / 2.0, i + 2.0, 1.0)
            # print(i, coeffs[i],lgam)
            for i in range(i - 1, 0, -1):
                lgam = coeffs[i] - x * lgam
                # print(lgam)
            coeffs0Minus1Third = 0
            FiveOver3Minusln3Minuseulers_const = 0
            return (
                (
                    x * x * (coeffs0Minus1Third - x * lgam)
                    - (x + 2.5) * log1(x / 3.0)
                    + log1(x / 2.0)
                )
                + FiveOver3Minusln3Minuseulers_const * x
            ) + lfb_2
        else:
            x = x - 3.0
            i = len(coeffs) - 1
            lgam = coeffs[i] * logcf(-x / 2.0, i + 2.0, 1.0)
            # print(i, coeffs[i],lgam)
            for i in range(i - 1, 0, -1):
                lgam = coeffs[i] - x * lgam
                # print(lgam)
            Forty7Over48Minusln4Minuseulers_const = 0
            return (
                (
                    x * x * (coeffs0Minusp25 - x * lgam)
                    - (x + 3.5) * log1(x / 4.0)
                    + log1(x / 2.0)
                    + log1(x / 3.0)
                )
                + Forty7Over48Minusln4Minuseulers_const * x
            ) + lfb_3

    elif x > -1.0:
        return logfbitdif(x) + logfbit(x + 1.0)
    else:
        raise ValueError


def lfbaccdif1(a, b):
    # Calculates logfbit(b)-logfbit(a+b) accurately for a > 0 & b >= 0.
    # Reasonably accurate for a >=0 & b < 0.
    if a < 0.0:
        return -lfbaccdif1(-a, b + a)
    elif b >= 8.0:
        y1 = b + 1.0
        y2 = y1**-2
        x1 = a + b + 1.0
        x2 = x1**-2
        # x3 = x2 * lfbc9
        y3 = y2 * lfbc9
        acc = x2 * (a * (x1 + y1) * y3)
        # x3 = x2 * (lfbc8 - x3)
        y3 = y2 * (lfbc8 - y3)
        acc = x2 * (a * (x1 + y1) * y3 - acc)
        # x3 = x2 * (lfbc7 - x3)
        y3 = y2 * (lfbc7 - y3)
        acc = x2 * (a * (x1 + y1) * y3 - acc)
        # x3 = x2 * (lfbc6 - x3)
        y3 = y2 * (lfbc6 - y3)
        acc = x2 * (a * (x1 + y1) * y3 - acc)
        # x3 = x2 * (lfbc5 - x3)
        y3 = y2 * (lfbc5 - y3)
        acc = x2 * (a * (x1 + y1) * y3 - acc)
        # x3 = x2 * (lfbc4 - x3)
        y3 = y2 * (lfbc4 - y3)
        acc = x2 * (a * (x1 + y1) * y3 - acc)
        # x3 = x2 * (lfbc3 - x3)
        y3 = y2 * (lfbc3 - y3)
        acc = x2 * (a * (x1 + y1) * y3 - acc)
        # x3 = x2 * (lfbc2 - x3)
        y3 = y2 * (lfbc2 - y3)
        acc = x2 * (a * (x1 + y1) * y3 - acc)
        # return lfbc1 * (a * (1.0 - y3) - y1 * acc) / (x1 * y1)
        return (a * (1.0 - y3) - y1 * acc) / (12.0 * x1 * y1)
    elif b >= 1.7:
        y1 = b + 1.0
        x1 = a + b + 1.0
        if b >= 3.0:
            Start = 17
        else:
            Start = 29

        s1 = (0.5 * (Start + 1.0)) ** 2
        s2 = (0.5 * (Start + 1.5)) ** 2
        ty = y1 * math.sqrt(1.0 + s1 * (y1**-2))
        tx = x1 * math.sqrt(1.0 + s1 * (x1**-2))
        y2 = ty - y1
        x2 = tx - x1
        acc = a * (1.0 - (2.0 * y1 + a) / (tx + ty))
        # Seems to work better without the next 2 lines. - Not with modification to s2
        ty = y1 * math.sqrt(1.0 + s2 * (y1**-2))
        tx = x1 * math.sqrt(1.0 + s2 * (x1**-2))
        acc = 0.25 * (
            acc + s1 / ((y1 + ty) * (x1 + tx)) * a * (1.0 + (2.0 * y1 + a) / (tx + ty))
        )
        y2 = 0.25 * (y2 + s1 / (y1 + ty))
        x2 = 0.25 * (x2 + s1 / (x1 + tx))
        for i in range(Start, 0, -1):
            acc = lfbArray[i] * (a - acc) / ((x1 + x2) * (y1 + y2))
            y2 = lfbArray[i] / (y1 + y2)
            x2 = lfbArray[i] / (x1 + x2)
        return cf_0 * (a - acc) / ((x1 + x2) * (y1 + y2))
        # return (a - acc) / (12.0 * (x1 + x2) * (y1 + y2))
    elif b > -1.0:
        if b < -0.66:
            if a > 1.0:
                return logfbitdif(b) + lfbaccdif1(a - 1.0, b + 1.0)
            elif a == 1.0:
                return logfbitdif(b)
            else:
                s2 = a * log0(1.0 / (b + 1.0 + a))
                s1 = logfbitdif(b + a)
                if s1 > s2:
                    s1 = (b + 1.5) * log0(a / ((b + 1.0) * (b + 2.0 + a))) - s2
                else:
                    s2 = s1
                    s1 = logfbitdif(b) - s1

                if s1 > 0.1 * s2:
                    return s1 + lfbaccdif1(a, b + 1.0)

        if b + a > 2:
            s1 = lfbaccdif1(b + a - 1.75, 1.75)
            a = 1.75 - b
        else:
            s1 = 0.0

        y1 = b - 1.0
        x1 = y1 + a
        i = len(coeffs2) + 2
        scale2 = 2.0**-i
        scale3 = 3.0**-i
        y2 = (
            (y1 + 2.5) * logcf(-y1 / 2.0, i, 1.0) - (2.0 / (i - 1.0))
        ) * scale2 + scale3 * logcf(-y1 / 3.0, i, 1.0)
        x2 = (
            (x1 + 2.5) * logcf(-x1 / 2.0, i, 1.0) - (2.0 / (i - 1.0))
        ) * scale2 + scale3 * logcf(-x1 / 3.0, i, 1.0)
        if a > 0.000006:
            acc = (
                y2 - x2
            )  # This calculation is not accurate enough for b < 0 and a small -
            # hence if b < 0 code above and derivative code below for small a
        else:
            y3 = -(y1 + a / 2.0) / 2.0
            x3 = -(y1 + a / 2.0) / 3.0
            acc = -a * (
                scale2
                * (
                    logcf(y3, i, 1.0)
                    + (y3 - 1.25) * (1.0 / (1.0 - y3) - i * logcf(y3, i + 1.0, 1.0))
                )
                - scale3 / 3.0 * (1.0 / (1.0 - x3) - i * logcf(x3, i + 1.0, 1.0))
            )

        for i in range(i - 3, -1, -1):
            acc = a * y2 - x1 * acc
            y2 = coeffs2[i] - y1 * y2
            x2 = coeffs2[i] - x1 * x2
        return s1 + (
            y1 * y1 * acc - a * (x2 * (x1 + y1) + Onep25Minusln2Minuseulers_const)
        )
    else:
        raise ValueError


def hypergeometricTerm(ai, aji, aki, amkji):
    # Probability that hypergeometric variate from a population with total type Is of
    # aki+ai, total type IIs of amkji+aji, has ai type Is and aji type IIs selected.
    # Parameterised this way for use with the Beta Negative Binomial distribution.
    ak = aki + ai
    amk = amkji + aji
    aj = aji + ai
    am = amk + ak
    amj = amkji + aki
    if am > max_discrete:
        raise ValueError
    if (ai == 0.0) and ((aji <= 0.0) or (aki <= 0.0) or (amj < 0.0) or (amk < 0.0)):
        return 1.0
    elif (ai > 0.0) and (min(aki, aji) == 0.0) and (max(amj, amk) == 0.0):
        return 1.0
    elif (ai >= 0.0) and (amkji > -1.0) and (aki > -1.0) and (aji >= 0.0):
        c1 = logfbit(amkji) + logfbit(aki) + logfbit(aji) + logfbit(am) + logfbit(ai)
        c1 = logfbit(amk) + logfbit(ak) + logfbit(aj) + logfbit(amj) - c1
        ai1 = ai + 1.0
        aj1 = aj + 1.0
        ak1 = ak + 1.0
        am1 = am + 1.0
        aki1 = aki + 1.0
        aji1 = aji + 1.0
        amk1 = amk + 1.0
        amj1 = amj + 1.0
        amkji1 = amkji + 1.0
        cjkmi = Generalabminuscd(aji, aki, ai, amkji)
        c5 = (cjkmi - ai) / (amkji1 * am1)
        if c5 < minLog1Value:
            c3 = amkji * (math.log((amj1 * amk1) / (amkji1 * am1)) - c5) - c5
        else:
            c3 = amkji * log1(c5) - c5

        c5 = (-cjkmi - aji) / (aki1 * am1)
        if c5 < minLog1Value:
            c4 = aki * (math.log((ak1 * amj1) / (aki1 * am1)) - c5) - c5
        else:
            c4 = aki * log1(c5) - c5

        c3 = c3 + c4
        c5 = (-cjkmi - aki) / (aji1 * am1)
        if c5 < minLog1Value:
            c4 = aji * (math.log((aj1 * amk1) / (aji1 * am1)) - c5) - c5
        else:
            c4 = aji * log1(c5) - c5

        c3 = c3 + c4
        c5 = (cjkmi - amkji) / (ai1 * am1)
        if c5 < minLog1Value:
            c4 = ai * (math.log((aj1 * ak1) / (ai1 * am1)) - c5) - c5
        else:
            c4 = ai * log1(c5) - c5

        c3 = c3 + c4

        return (
            math.exp((c1 + 1.0 / am1) + c3)
            * math.sqrt(
                (amk1 * ak1) * (aj1 * amj1) / ((amkji1 * aki1 * aji1) * (am1 * ai1))
            )
            * OneOverSqrTwoPi
        )
    else:
        return 0.0


def hypergeometric(ai, aji, aki, amkji, comp):  # , ha1, hprob, hswap):
    # Probability that hypergeometric variate from a population with total type Is of
    # aki+ai, total type IIs of amkji+aji, has up to ai type Is selected in a sample
    # of size aji+ai.
    # Parameterised this way for use with the Beta Negative Binomial distribution.
    global ha1, hprob, hswap
    if (amkji > -1.0) and (amkji < 0.0):
        ip1 = -amkji
        mkji = ip1 - 1.0
        allintegral = False
    else:
        ip1 = amkji + 1.0
        mkji = amkji
        allintegral = (
            ai == math.floor(ai)
            and aji == math.floor(aji)
            and aki == math.floor(aki)
            and mkji == math.floor(mkji)
        )

    if allintegral:
        swapped = (ai + 0.5) * (mkji + 0.5) >= (aki - 0.5) * (aji - 0.5)
    elif ai < 100.0 and ai == math.floor(ai) or mkji < 0.0:
        if comp:
            swapped = (ai + 0.5) * (mkji + 0.5) >= aki * aji
        else:
            swapped = (ai + 0.5) * (mkji + 0.5) >= aki * aji + 1000.0

    elif ai < 1.0:
        swapped = (ai + 0.5) * (mkji + 0.5) >= aki * aji
    elif aji < 1.0 or aki < 1.0 or (ai < 1.0 and ai > 0.0):
        swapped = False
    else:
        swapped = (ai + 0.5) * (mkji + 0.5) >= (aki - 0.5) * (aji - 0.5)

    if not swapped:
        i = ai
        ji = aji
        ki = aki
    else:
        i = aji - 1.0
        ji = ai + 1.0
        ki = ip1
        ip1 = aki
        mkji = aki - 1.0

    c2 = ji + i
    c4 = mkji + ki + c2
    if c4 > max_discrete:
        raise ValueError

    if (
        (i >= 0.0)
        and (ji <= 0.0)
        or (ki <= 0.0)
        or (ip1 + ki <= 0.0)
        or (ip1 + ji <= 0.0)
    ):
        exact = True
        if i >= 0.0:
            prob = 1.0
        else:
            prob = 0.0

    elif (ip1 > 0.0) and (ip1 < 1.0):
        exact = False
        prob = (
            hypergeometricTerm(i, ji, ki, ip1)
            * (ip1 * (c4 + 1.0))
            / ((ki + ip1) * (ji + ip1))
        )
    else:
        exact = (
            (i == 0.0)
            and ((ji <= 0.0) or (ki <= 0.0) or (mkji + ki < 0.0) or (mkji + ji < 0.0))
        ) or ((i > 0.0) and (min(ki, ji) == 0.0) and (max(mkji + ki, mkji + ji) == 0.0))
        prob = hypergeometricTerm(i, ji, ki, mkji)

    hprob = prob
    hswap = swapped
    ha1 = 0.0

    if (exact) or (prob == 0.0):
        if swapped == comp:
            return prob
        else:
            return 1.0 - prob

    a1 = 0.0
    sumAlways = 0.0
    sumFactor = 10.0

    if i < mkji:
        must_do_cf = i != math.floor(i)
        maxSums = math.floor(i)
    else:
        must_do_cf = mkji != math.floor(mkji)
        maxSums = math.floor(max(mkji, 0.0))

    if must_do_cf:
        sumAlways = 0.0
        sumFactor = 5.0
    else:
        sumAlways = 20.0
        sumFactor = 10.0

    if maxSums > sumAlways or must_do_cf:
        numb = math.floor(
            sumFactor
            / c4
            * math.exp(math.log((ki + i) * (ji + i) * (ip1 + ji) * (ip1 + ki)) / 3.0)
        )
        numb = math.floor(i - (ki + i) * (ji + i) / c4 + numb)
        if numb < 0.0:
            numb = 0.0
        elif numb > maxSums:
            numb = maxSums
    else:
        numb = maxSums

    if 2.0 * numb <= maxSums or must_do_cf:
        b1 = 1.0
        c1 = 0.0
        c2 = i - numb
        c3 = mkji - numb
        s = c3
        a2 = c2
        c3 = c3 - 1.0
        b2 = Generalabminuscd(ki + numb + 1.0, ji + numb + 1.0, c2 - 1.0, c3)
        bn = b2
        bnAdd = c3 + c4 + c2 - 2.0
        while b2 > 0.0 and (abs(a2 * b1 - a1 * b2) > abs(cfVSmall * b1 * a2)):
            c1 = c1 + 1.0
            c2 = c2 - 1.0
            an = (c1 * c2) * (c3 * c4)
            c3 = c3 - 1.0
            c4 = c4 - 1.0
            bn = bn + bnAdd
            bnAdd = bnAdd - 4.0
            a1 = bn * a2 + an * a1
            b1 = bn * b2 + an * b1
            if b1 > scalefactor:
                a1 = a1 * scalefactor2
                b1 = b1 * scalefactor2
                a2 = a2 * scalefactor2
                b2 = b2 * scalefactor2
            c1 = c1 + 1.0
            c2 = c2 - 1.0
            an = (c1 * c2) * (c3 * c4)
            c3 = c3 - 1.0
            c4 = c4 - 1.0
            bn = bn + bnAdd
            bnAdd = bnAdd - 4.0
            a2 = bn * a1 + an * a2
            b2 = bn * b1 + an * b2
            if b2 > scalefactor:
                a1 = a1 * scalefactor2
                b1 = b1 * scalefactor2
                a2 = a2 * scalefactor2
                b2 = b2 * scalefactor2

        if b1 < 0.0 or b2 < 0.0:
            raise ValueError  # Actually, something unexpected has happened here -
            # although it will be to do with the values input.
        else:
            a1 = a2 / b2 * s

    else:
        numb = maxSums

    c1 = i - numb + 1.0
    c2 = mkji - numb + 1.0
    c3 = ki + numb
    c4 = ji + numb
    while numb > 0.0:
        a1 = (1.0 + a1) * ((c1 * c2) / (c3 * c4))
        c1 = c1 + 1.0
        c2 = c2 + 1.0
        c3 = c3 - 1.0
        c4 = c4 - 1.0
        numb = numb - 1.0

    ha1 = a1
    a1 = (1.0 + a1) * prob
    if swapped == comp:
        return a1
    else:
        if a1 > 0.99:
            raise ValueError  # Similarly, if we end up here then something unexpected
            # has happened - although it will be to do with the values input.
        else:
            return 1.0 - a1


def PBB(i, ssmi, beta_shape1, beta_shape2):
    global hTerm
    hTerm = hypergeometricTerm(i, ssmi, beta_shape1, beta_shape2)
    return (
        (beta_shape1 / (i + beta_shape1))
        * (beta_shape2 / (beta_shape1 + beta_shape2))
        * ((i + ssmi + beta_shape1 + beta_shape2) / (ssmi + beta_shape2))
        * hTerm
    )


def pmf_BetaNegativeBinomial(i, r, beta_shape1, beta_shape2):
    i = AlterForIntegralChecks_Others(i)
    if r <= 0.0 or beta_shape1 <= 0.0 or beta_shape2 <= 0.0:
        raise ValueError
    elif i < 0:
        return 0.0
    else:
        return (
            (beta_shape2 / (beta_shape1 + beta_shape2))
            * (r / (beta_shape1 + r))
            * beta_shape1
            * (i + beta_shape1 + r + beta_shape2)
            / ((i + r) * (i + beta_shape2))
            * hypergeometricTerm(i, r, beta_shape2, beta_shape1)
        )


def CBNB0(i, r, beta_shape1, beta_shape2, toBeAdded):
    global ha1, hprob, hswap
    if r < 2.0 or beta_shape2 < 2.0:
        # Assumption here that i is integral or greater than 4.
        mrb2 = max(r, beta_shape2)
        other = min(r, beta_shape2)
        cbnb0 = PBB(i, other, mrb2, beta_shape1)
        if i == 0.0:
            return cbnb0
        cbnb0 = cbnb0 * (
            1.0 + i * (other + beta_shape1) / (((i - 1.0) + mrb2) * (other + 1.0))
        )
        if i == 1.0:
            return cbnb0
        i = i - 2.0
        other = other + 2.0
        temp = PBB(i, mrb2, other, beta_shape1)
        if i == 0.0:
            return cbnb0 + temp
        cbnb0 = cbnb0 + temp * (
            1.0 + i * (mrb2 + beta_shape1) / (((i - 1.0) + other) * (mrb2 + 1.0))
        )
        if i == 1.0:
            return cbnb0
        i = i - 2.0
        mrb2 = mrb2 + 2.0
        return cbnb0 + CBNB0(i, mrb2, beta_shape1, other, cbnb0)
    elif beta_shape1 < 1.0:
        mrb2 = max(r, beta_shape2)
        other = min(r, beta_shape2)
        cbnb0 = hypergeometric(
            i, mrb2 - 1.0, other, beta_shape1, False
        )  # , ha1, hprob, hswap)
        if hswap:
            temp = PBB(mrb2 - 1.0, beta_shape1, i + 1.0, other)
            if (toBeAdded + (cbnb0 - temp)) < 0.01 * (toBeAdded + (cbnb0 + temp)):
                return CBNB2(i, mrb2, beta_shape1, other)
            else:
                return cbnb0 - temp

        elif ha1 < -0.9 * beta_shape1 / (beta_shape1 + other):
            raise ValueError
        else:
            return hprob * (beta_shape1 / (beta_shape1 + other) + ha1)
    else:
        return hypergeometric(
            i, r, beta_shape2, beta_shape1 - 1.0, False
        )  # , ha1, hprob, hswap)


def CBNB2(i, r, beta_shape1, beta_shape2):
    ss = min(r, beta_shape2)
    bs2 = max(r, beta_shape2)
    r = ss
    beta_shape2 = bs2
    d1 = (i + 0.5) * (beta_shape1 + 0.5) - (bs2 - 1.5) * (ss - 0.5)
    if d1 < 0.0:
        return CBNB0(i, ss, beta_shape1, bs2, 0.0)

    d1 = math.floor(d1 / (bs2 + beta_shape1 - 1.0)) + 10.0
    if ss + d1 > bs2:
        d1 = math.floor(bs2 - ss)
    ss = ss + d1
    j = i - d1
    d2 = (j + 0.5) * (beta_shape1 + 0.5) - (bs2 - 1.5) * (ss - 0.5)
    if d2 < 0.0:
        d2 = 10.0
    else:
        temp = bs2 + ss + 2.0 * beta_shape1 - 1.0
        d2 = math.floor((math.sqrt(temp * temp + 4.0 * d2) - temp) / 2.0) + 10.0

    if 2.0 * d2 > i:
        d2 = math.floor(i / 2.0)

    pbbval = PBB(i, r, beta_shape2, beta_shape1)
    ss = ss + d2
    bs2 = bs2 + d2
    j = j - 2.0 * d2
    cbnb2 = CBNB0(j, ss, beta_shape1, bs2, 0.0)
    temp = 1.0
    d_count = d2 - 2.0
    # print(d1,d2)
    j = j + 1.0
    while d_count >= 0.0:
        j = j + 1.0
        bs2 = beta_shape2 + d_count
        d_count = d_count - 1.0
        temp = 1.0 + (j * (bs2 + beta_shape1) / ((j + ss - 1.0) * (bs2 + 1.0))) * temp

    j = i - d2 - d1
    temp = (ss * (j + bs2)) / (bs2 * (j + ss)) * temp
    d_count = d1 + d2 - 1.0
    while d_count >= 0:
        j = j + 1.0
        ss = r + d_count
        d_count = d_count - 1.0
        temp = 1.0 + (j * (ss + beta_shape1) / ((j + bs2 - 1.0) * (ss + 1.0))) * temp

    return cbnb2 + temp * pbbval


def ccBNB5(ilim, rr, a, bb):
    if rr > bb:
        r = rr
        b = bb
    else:
        r = bb
        b = rr

    ccbnb5 = (
        (a + 0.5) * log0(b * r / ((a + 1.0) * (b + a + r + 1.0)))
        - r * log0(b / (r + a + 1.0))
        - b * log0(r / (a + b + 1.0))
    )
    if r <= 0.001:
        temp = a + (b + r) * 0.5
        ccbnb5 = ccbnb5 - b * r * (
            logfbit2(temp) + (b**2 + r**2) * logfbit4(temp) / 24.0
        )
    else:
        ccbnb5 = ccbnb5 + (lfbaccdif1(b, r + a) - lfbaccdif1(b, a))

    temp = 0.0
    if ilim > 0.0:
        i = ilim
        while i > 1.0:
            i = i - 1.0
            temp = (1.0 + temp) * (i + r) * (i + b) / ((i + r + a + b) * (i + 1.0))

        temp = (1.0 + temp) * math.exp(ccbnb5) * a

    return (r * b * (1.0 - temp) - math.expm1(ccbnb5) * a * (r + a + b)) / (
        (r + a) * (a + b)
    )


def cdf_BetaNegativeBinomial(i, r, beta_shape1, beta_shape2):
    i = math.floor(i)
    if r <= 0.0 or beta_shape1 <= 0.0 or beta_shape2 <= 0.0:
        raise ValueError
    elif i < 0:
        return 0.0
    else:
        return CBNB0(i, r, beta_shape1, beta_shape2, 0.0)


def sf_BetaNegativeBinomial(i, r, beta_shape1, beta_shape2):
    i = math.floor(i)
    mrb2 = max(r, beta_shape2)
    other = min(r, beta_shape2)
    if other <= 0.0 or beta_shape1 <= 0.0 or beta_shape2 <= 0.0:
        raise ValueError
    elif i < 0.0:
        return 1.0
    elif (i == 0.0) or (
        (i < 1000000.0)
        and (other < 0.001)
        and (beta_shape1 > 50.0 * other)
        and (10.0 * i * beta_shape1 < mrb2)
    ):
        return ccBNB5(i, mrb2, beta_shape1, other)
    elif (
        mrb2 >= 100.0
        or other > 20.0
        or (
            mrb2 >= 5.0
            and (other - 0.5) * (mrb2 - 0.5) > (i + 0.5) * (beta_shape1 + 0.5)
        )
    ):
        return CBNB0(mrb2 - 1.0, i + 1.0, other, beta_shape1, 0.0)
    else:
        ccbnb = 0.0
        temp = 0.0
        i = i + 1.0
        if other >= 1.0:
            mrb2 = mrb2 - 1.0
            other = other - 1.0
            temp = hypergeometricTerm(i, mrb2, other, beta_shape1)
            ccbnb = temp
            while (other >= 1.0) and (temp > 1e-16 * ccbnb):
                i = i + 1.0
                beta_shape1 = beta_shape1 + 1.0
                temp = temp * (mrb2 * other) / (i * beta_shape1)
                mrb2 = mrb2 - 1.0
                other = other - 1.0
                ccbnb = ccbnb + temp

            if other >= 1.0:
                return ccbnb
            i = i + 1.0
            beta_shape1 = beta_shape1 + 1.0

        if mrb2 >= 1.0:
            mxib1 = max(i, beta_shape1)
            mnib1 = min(i, beta_shape1)
            if temp == 0.0:
                mrb2 = mrb2 - 1.0
                temp = PBB(mnib1, mrb2, other, mxib1)
            else:  # temp is hypergeometricTerm(mnib1-1, mrb2, other, mxib1-1)
                temp = temp * other * mrb2
                mrb2 = mrb2 - 1.0
                temp = temp / (mnib1 * (mrb2 + mxib1))

            ccbnb = ccbnb + temp
            while (mrb2 >= 1.0) and (temp > 1e-16 * ccbnb):
                temp = temp * mrb2 * (mnib1 + other)
                mnib1 = mnib1 + 1.0
                if mnib1 > mxib1:
                    mxib1, mnib1 = mnib1, mxib1

                # Block below not required if hypergeometric block included above and
                # therefore other guaranteed < 1 <= mrb2
                # if mrb2 < other:
                #    other, mrb2 = mrb2, other
                #
                mrb2 = mrb2 - 1.0
                temp = temp / ((mrb2 + mxib1) * mnib1)
                ccbnb = ccbnb + temp

            if mrb2 >= 1.0:
                return ccbnb
            temp = temp * mrb2 / (mnib1 + mrb2)
        else:
            mxib1 = beta_shape1
            mnib1 = i
            if temp == 0.0:
                temp = pmf_BetaNegativeBinomial(mnib1, mrb2, mxib1, other)
            else:
                temp = temp * mrb2 * other / (i * (mrb2 + other + mnib1 + mxib1 + -1))

            ccbnb = ccbnb + temp

        max_iterations = 60.0
        while True:
            temp = (
                temp * (mnib1 + mrb2) * (mnib1 + other) / (mnib1 + mxib1 + mrb2 + other)
            )
            mnib1 = mnib1 + 1.0
            if mxib1 < mnib1:
                mxib1, mnib1 = mnib1, mxib1

            temp = temp / mnib1
            ccbnb = ccbnb + temp
            if (temp <= 1e-16 * ccbnb) or (mnib1 + mxib1 > max_iterations):
                break

        temp = temp * (mnib1 + mrb2) * (mnib1 + other) / ((mnib1 + 1.0) * mxib1)
        ccbnb = ccbnb + temp
        mnib1 = mnib1 + 1.0
        mrb2 = mrb2 - 1.0
        other = other - 1.0
        while True:
            mnib1 = mnib1 + 1.0
            mxib1 = mxib1 + 1.0
            temp = temp * (mrb2 * other) / (mnib1 * mxib1)
            mrb2 = mrb2 - 1.0
            other = other - 1.0
            ccbnb = ccbnb + temp
            if abs(temp) <= 1e-16 * ccbnb:
                break

    return ccbnb


def pmf_BetaBinomial(i, sample_size, beta_shape1, beta_shape2):
    i = AlterForIntegralChecks_Others(i)
    sample_size = AlterForIntegralChecks_Others(sample_size)
    # These probably should be checked once, when creating a BetaBinomial distribution
    # object
    if (beta_shape1 <= 0.0) or (beta_shape2 <= 0.0) or (sample_size < 0.0):
        raise ValueError
    elif (i < 0) or (i > sample_size):
        return 0.0
    else:
        return (
            (beta_shape1 / (i + beta_shape1))
            * (beta_shape2 / (beta_shape1 + beta_shape2))
            * (
                (sample_size + beta_shape1 + beta_shape2)
                / (sample_size - i + beta_shape2)
            )
            * hypergeometricTerm(i, sample_size - i, beta_shape1, beta_shape2)
        )


def cdf_BetaBinomial(i, sample_size, beta_shape1, beta_shape2):
    i = math.floor(i)
    sample_size = AlterForIntegralChecks_Others(sample_size)
    if (beta_shape1 <= 0.0) or (beta_shape2 <= 0.0) or (sample_size < 0.0):
        raise ValueError
    elif i < 0.0:
        return 0.0
    else:
        i = i + 1.0
        return sf_BetaNegativeBinomial(sample_size - i, i, beta_shape1, beta_shape2)


def sf_BetaBinomial(i, sample_size, beta_shape1, beta_shape2):
    i = math.floor(i)
    sample_size = AlterForIntegralChecks_Others(sample_size)
    if (beta_shape1 <= 0.0) or (beta_shape2 <= 0.0) or (sample_size < 0.0):
        raise ValueError
    elif i < 0.0:
        return 1.0
    elif i >= sample_size:
        return 0.0
    else:
        return sf_BetaNegativeBinomial(i, sample_size - i, beta_shape2, beta_shape1)


# Hypergeometric & Negative Hpergeometric pmfs, cdfs & sfs for completeness
def pmf_hypergeometric(type1s, sample_size, tot_type1, pop_size):
    type1s = AlterForIntegralChecks_Others(type1s)
    sample_size = AlterForIntegralChecks_Others(sample_size)
    tot_type1 = AlterForIntegralChecks_Others(tot_type1)
    pop_size = AlterForIntegralChecks_Others(pop_size)
    if (
        (sample_size < 0.0)
        or (tot_type1 < 0.0)
        or (sample_size > pop_size)
        or (tot_type1 > pop_size)
    ):
        raise ValueError
    else:
        return hypergeometricTerm(
            type1s,
            sample_size - type1s,
            tot_type1 - type1s,
            pop_size - tot_type1 - sample_size + type1s,
        )


def cdf_hypergeometric(type1s, sample_size, tot_type1, pop_size):
    type1s = math.floor(type1s)
    sample_size = AlterForIntegralChecks_Others(sample_size)
    tot_type1 = AlterForIntegralChecks_Others(tot_type1)
    pop_size = AlterForIntegralChecks_Others(pop_size)
    if (
        (sample_size < 0.0)
        or (tot_type1 < 0.0)
        or (sample_size > pop_size)
        or (tot_type1 > pop_size)
    ):
        raise ValueError
    else:
        return hypergeometric(
            type1s,
            sample_size - type1s,
            tot_type1 - type1s,
            pop_size - tot_type1 - sample_size + type1s,
            False,
        )


def sf_hypergeometric(type1s, sample_size, tot_type1, pop_size):
    type1s = math.floor(type1s)
    sample_size = AlterForIntegralChecks_Others(sample_size)
    tot_type1 = AlterForIntegralChecks_Others(tot_type1)
    pop_size = AlterForIntegralChecks_Others(pop_size)
    if (
        (sample_size < 0.0)
        or (tot_type1 < 0.0)
        or (sample_size > pop_size)
        or (tot_type1 > pop_size)
    ):
        raise ValueError
    else:
        return hypergeometric(
            type1s,
            sample_size - type1s,
            tot_type1 - type1s,
            pop_size - tot_type1 - sample_size + type1s,
            True,
        )


def pmf_neghypergeometric(type2s, type1s_reqd, tot_type1, pop_size):
    type2s = AlterForIntegralChecks_Others(type2s)
    type1s_reqd = AlterForIntegralChecks_Others(type1s_reqd)
    tot_type1 = AlterForIntegralChecks_Others(tot_type1)
    pop_size = AlterForIntegralChecks_Others(pop_size)
    if (type1s_reqd <= 0.0) or (tot_type1 < type1s_reqd) or (tot_type1 > pop_size):
        raise ValueError
    elif (type2s < 0.0) or (tot_type1 + type2s > pop_size):
        if type2s == 0.0:
            return 1.0
        else:
            return 0.0
    else:
        return (
            hypergeometricTerm(
                type1s_reqd - 1.0,
                type2s,
                tot_type1 - type1s_reqd + 1.0,
                pop_size - tot_type1 - type2s,
            )
            * (tot_type1 - type1s_reqd + 1.0)
            / (pop_size - type1s_reqd - type2s + 1.0)
        )


def cdf_neghypergeometric(type2s, type1s_reqd, tot_type1, pop_size):
    type2s = math.floor(type2s)
    type1s_reqd = AlterForIntegralChecks_Others(type1s_reqd)
    tot_type1 = AlterForIntegralChecks_Others(tot_type1)
    pop_size = AlterForIntegralChecks_Others(pop_size)
    if (type1s_reqd <= 0.0) or (tot_type1 < type1s_reqd) or (tot_type1 > pop_size):
        raise ValueError
    elif tot_type1 + type2s > pop_size:
        return 1.0
    else:
        return hypergeometric(
            type2s,
            type1s_reqd,
            pop_size - tot_type1 - type2s,
            tot_type1 - type1s_reqd,
            False,
        )


def sf_neghypergeometric(type2s, type1s_reqd, tot_type1, pop_size):
    type2s = math.floor(type2s)
    type1s_reqd = AlterForIntegralChecks_Others(type1s_reqd)
    tot_type1 = AlterForIntegralChecks_Others(tot_type1)
    pop_size = AlterForIntegralChecks_Others(pop_size)
    if (type1s_reqd <= 0.0) or (tot_type1 < type1s_reqd) or (tot_type1 > pop_size):
        raise ValueError
    elif tot_type1 + type2s > pop_size:
        return 0.0
    else:
        return hypergeometric(
            type2s,
            type1s_reqd,
            pop_size - tot_type1 - type2s,
            tot_type1 - type1s_reqd,
            True,
        )
